# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
import numpy as np


def neg(x):
    return -x


def pos(x):
    return x


# def abs(x):
#     return np.abs(x)


def not_func(x):
    return (~(x != 0)).astype(int)


def add(x, y):
    return x + y


def sub(x, y):
    return x - y


def mul(x, y):
    return x * y


def div(x, y):
    return x / y


def pow(x, y):
    return x**y


def ge(x, y):
    return (x >= y).astype(int)


def gt(x, y):
    return (x > y).astype(int)


def le(x, y):
    return (x <= y).astype(int)


def lt(x, y):
    return (x < y).astype(int)


def eq(x, y):
    return (x == y).astype(int)


def ne(x, y):
    return (x != y).astype(int)


def and_func(x, y):
    return ne(x, 0) * ne(y, 0)


def or_func(x, y):
    return np.clip(ne(x, 0) + ne(y, 0), 0, 1)


def set_from_other(x, y):
    return y


# single argument functions


def abs(x):
    return np.fabs(x)


def acos(x):
    return np.arccos(x)


def asin(x):
    return np.arcsin(x)


def atan(x):
    return np.arctan(x)


def cos(x):
    return np.cos(x)


def exp(x):
    return np.exp(x)


def log(x):
    return np.log(x)


def log10(x):
    return np.log10(x)


def sgn(x):
    return np.sign(x)


def square(x):
    return np.square(x)


def sqrt(x):
    return np.sqrt(x)


def sin(x):
    return np.sin(x)


def tan(x):
    return np.tan(x)


# double argument functions


def atan2(x, y):
    return np.arctan2(x, y)


def floor_div(x, y):
    return np.floor_divide(x, y)


def mod(x, y):
    return np.mod(x, y)


# bitmapping


def bitmap(x, y):
    if isinstance(y, (int, float)):
        x[x == y] = np.nan
        return x
    elif isinstance(y, np.ndarray):
        x[np.isnan(y)] = np.nan
        return x


def nobitmap(x, y):
    x[np.isnan(x)] = y
    return x



#
# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
import json
import logging
import os
from pathlib import Path

import yaml
import metview as mv
from metview.metviewpy.param import ParamInfo
from metview.metviewpy.ipython import is_ipython_active
from metview.scaling import Scaling

# logging.basicConfig(level=logging.DEBUG)
LOG = logging.getLogger(__name__)

_DB = {
    "param": [None, "params.yaml", "param_styles.yaml"],
    "map": [None, "", "map_styles.yaml"],
}

_MAP_CONF = None
ETC_PATH = os.path.join(os.path.dirname(__file__), "etc")
CUSTOM_CONF_PATH = []
LOCAL_CONF_PATH = ""


PARAM_VISDEF_VERBS = ["mcont", "mwind", "mcoast", "msymb", "mgraph"]


# _DB = None


# def cont():
#     global _DB
#     if not _DB:
#         _DB = StyleDb()
#     return _DB


class ContourStyleDbItem:
    def __init__(self, name, db):
        self.name = name
        self.db = db
        self.keywords = []
        self.colours = []
        self.layers = []
        self.description = ""

    def list_match(self, lst, pattern):
        p = pattern.lower()
        for t in lst:
            if p in t.lower():
                return True
        return False

    def keyword_match(self, pattern):
        return self.list_match(self.keywords, pattern)

    def layer_match(self, pattern):
        return self.list_match(self.layers, pattern)

    def colour_match(self, pattern):
        return self.list_match(self.colours, pattern)

    def preview_file(self):
        if self.db.preview_path:
            return os.path.join(self.db.preview_path, self.name + ".png")
        else:
            return str()


class ContourStyleDb:
    def __init__(self):
        self.SHARE_DIR = os.path.join(
            mv.version_info()["metview_dir"], "share", "metview", "eccharts"
        )
        self.preview_path = os.path.join(self.SHARE_DIR, "style_previews")
        self.items = []
        self.keywords = []
        self.colours = []
        self.layers = []
        self._load()
        self._load_magics_def()

    def _load(self):
        file_path = os.path.join(self.SHARE_DIR, "styles.json")
        # print(f"file_path={file_path}")
        try:
            with open(file_path, "rt") as f:
                conf = json.load(f)
                col_set = set()
                key_set = set()
                layer_set = set()
                for name, c in conf.items():
                    item = ContourStyleDbItem(name, self)
                    item.colours = c.get("colours", [])
                    item.keywords = c.get("keywords", [])
                    item.layers = c.get("layers", [])
                    col_set.update(item.colours)
                    key_set.update(item.keywords)
                    layer_set.update(item.layers)
                    self.items.append(item)
            self.colours = sorted(list(col_set))
            self.keywords = sorted(list(key_set))
            self.layers = sorted(list(layer_set))
        except:
            pass
        #     LOG.exception("Failed to read eccharts styles", popup=True)

        self._load_magics_def()

    def find_by_name(self, name):
        for i, item in enumerate(self.items):
            if item.name == name:
                return i, item
        return -1, None

    def _load_magics_def(self):
        pass

    def names(self):
        return sorted([item.name for item in self.items])


class Visdef:
    # BUILDER = {
    #     "mcont": mv.mcont,
    #     "mwind": mv.mwind,
    #     "mcoast": mv.mcoast,
    #     "msymb": mv.msymb,
    #     "mgraph": mv.mgraph
    # }

    def __init__(self, verb, params):
        self.verb = verb.lower()
        self.params = params

        self.BUILDER = {
            "mcont": mv.mcont,
            "mwind": mv.mwind,
            "mcoast": mv.mcoast,
            "msymb": mv.msymb,
            "mgraph": mv.mgraph,
        }

    def clone(self):
        return Visdef(self.verb, copy.deepcopy(self.params))

    def change(self, verb, param, value):
        if verb == self.verb:
            self.params[param] = value

    def change_symbol_text_list(self, text_value, idx_value):
        assert self.verb == "msymb"
        if self.verb == "msymb":
            if self.params.get("symbol_type", "").lower() == "text":
                self.params["symbol_advanced_table_text_list"] = text_value
                self.params["symbol_advanced_table_level_list"] = idx_value

    def set_data_id(self, data_id):
        if self.verb in ["mcont", "mwind"] and data_id is not None and data_id != "":
            self.params["grib_id"] = data_id

    def set_values_list(self, values):
        if self.verb == "mcont":
            if self.params.get("contour_level_list", []):
                self.params["contour_level_list"] = values

    @staticmethod
    def from_request(req):
        params = {k: v for k, v in req.items() if not k.startswith("_")}
        vd = Visdef(req.verb.lower(), copy.deepcopy(params))
        return vd

    def to_request(self):
        fn = self.BUILDER.get(self.verb, None)
        if fn is not None:
            return fn(**(self.params))
        else:
            raise Exception(f"{self} unsupported verb!")

    def __str__(self):
        return f"Visdef[verb={self.verb}, params={self.params}]"

    def __repr__(self):
        return f"Visdef(verb={self.verb}, params={self.params})"


class Style:
    def __init__(self, name, visdefs):
        self.name = name
        self.visdefs = visdefs
        if not isinstance(visdefs, list):
            self.visdefs = [self.visdefs]

    def clone(self):
        return Style(self.name, [vd.clone() for vd in self.visdefs])

    def to_request(self):
        return [vd.to_request() for vd in self.visdefs]

    def update(self, *args, inplace=False, verb=None):
        s = self if inplace == True else self.clone()
        if isinstance(verb, str):
            verb = [verb]
        if verb:
            for v in args:
                if isinstance(v, dict):
                    v_vals = {v_key.lower(): v_val for v_key, v_val in v.items()}
                    for i, vd in enumerate(s.visdefs):
                        if vd.verb in verb:
                            vd.params.update(v_vals)
        else:
            for i, v in enumerate(args):
                if isinstance(v, dict) and i < len(s.visdefs):
                    v = {v_key.lower(): v_val for v_key, v_val in v.items()}
                    s.visdefs[i].params.update(v)
        return s

    def set_data_id(self, data_id):
        allowed_verbs = ["mcont", "mwind"]
        if isinstance(data_id, str) and any(x in allowed_verbs for x in self.verbs()):
            return self.update({"grib_id": data_id}, verb=allowed_verbs)
        else:
            return self

    def verbs(self):
        return [vd.verb for vd in self.visdefs]

    def __str__(self):
        t = f"{self.__class__.__name__}[name={self.name}] "
        for vd in self.visdefs:
            t += f"{vd} "
        return t


class ParamMatchCondition:
    def __init__(self, cond):
        self.name = cond.pop("info_name", "")
        self.cond = cond
        if "levelist" in self.cond:
            if not isinstance(self.cond["levelist"], list):
                self.cond["levelist"] = [self.cond["levelist"]]
        # ParamInfo use "typeOfLevel" instead of "levtype" so we need to
        # change the key a remap values
        if "levtype" in self.cond:
            v = self.cond.pop("levtype")
            self.cond["typeOfLevel"] = v
            self.cond["typeOfLevel"] = ParamInfo.LEVEL_TYPES.get(v, v)

        for k, v in self.cond.items():
            if isinstance(v, list):
                self.cond[k] = [str(x) for x in v]
            else:
                self.cond[k] = str(v)

    def match(self, param_info):
        return param_info.match(self.name, self.cond)


class ParamStyle:
    def __init__(self, conf, db):
        self.cond = []
        for d in conf["match"]:
            if "info_name" in d:
                self.info_name = d["info_name"]
            self.cond.append(ParamMatchCondition(d))
        self.param_type = conf.get("param_type", "scalar")

        if self.param_type == "vector":
            default_style = db.VECTOR_DEFAULT_STYLE_NAME
        else:
            default_style = db.SCALAR_DEFAULT_STYLE_NAME

        s = conf.get("styles", {})
        self.style = s.get("basic", [default_style])
        self.xs_style = s.get("xs", self.style)
        self.diff_style = s.get("diff", [db.DIFF_DEFAULT_STYLE_NAME])

        self.scaling = conf.get("scaling", [])

    def match(self, param):
        return max([d.match(param) for d in self.cond])

    def find_style(self, plot_type):
        if plot_type == "" or plot_type == "map":
            return self.style[0]
        elif plot_type == "diff":
            return self.diff_style[0]
        elif plot_type == "xs":
            return self.xs_style[0]
        else:
            return None

    def styles(self, plot_type):
        if plot_type == "" or plot_type == "map":
            return self.style
        elif plot_type == "diff":
            return self.diff_style
        elif plot_type == "xs":
            return self.xs_style
        else:
            return []

    def __str__(self):
        return "{}[name={},type={}]".format(
            self.__class__.__name__, self.info_name, self.param_type
        )


class StyleDb:
    SCALAR_DEFAULT_STYLE_NAME = "default_mcont"
    VECTOR_DEFAULT_STYLE_NAME = "default_mwind"
    DIFF_DEFAULT_STYLE_NAME = "default_diff"

    def __init__(self, param_file_name, style_file_name):
        self.params = []
        self.styles = {}
        self.param_file_name = param_file_name
        self.style_file_name = style_file_name

        self._load_system_config()
        self._load_custom_config()
        self._load_local_config()

    def _load_system_config(self):
        self._load(
            os.path.join(ETC_PATH, self.param_file_name)
            if self.param_file_name
            else "",
            os.path.join(ETC_PATH, self.style_file_name),
        )

    def _load_custom_config(self):
        if CUSTOM_CONF_PATH and CUSTOM_CONF_PATH[-1]:
            self._load(
                os.path.join(CUSTOM_CONF_PATH[-1], self.param_file_name)
                if self.param_file_name
                else "",
                os.path.join(CUSTOM_CONF_PATH[-1], self.style_file_name),
            )

    def _load_local_config(self):
        if LOCAL_CONF_PATH:
            self._load(
                os.path.join(LOCAL_CONF_PATH, self.param_file_name)
                if self.param_file_name
                else "",
                os.path.join(LOCAL_CONF_PATH, self.style_file_name),
            )

    def get_style(self, style):
        if style in self.styles:
            return self.styles[style]
        else:
            return self.styles.get("default", None)

    def _best_param_match(self, param_info):
        r = 0
        p_best = None
        for p in self.params:
            m = p.match(param_info)
            # print(f"p={p} m={m}")
            if m > r:
                r = m
                p_best = p
        return p_best

    def get_param_style_list(self, param_info, scalar=True, plot_type="map"):
        p_best = self._best_param_match(param_info)
        s = []
        # print(f"param_info={param_info}")
        if p_best is not None:
            s = p_best.styles(plot_type)
            # print(f" -> style_name={style_name}")
        if scalar:
            s.append(self.SCALAR_DEFAULT_STYLE_NAME)
        else:
            s.append(self.VECTOR_DEFAULT_STYLE_NAME)
        return s

    def get_param_style(self, param_info, scalar=True, plot_type="map", data_id=None):
        p_best = self._best_param_match(param_info)
        s = None
        # print(f"param_info={param_info}")
        if p_best is not None:
            style_name = p_best.find_style(plot_type)
            # print(f" -> style_name={style_name}")
            s = self.styles.get(style_name, None)
        else:
            if scalar:
                s = self.styles.get(self.SCALAR_DEFAULT_STYLE_NAME, None)
            else:
                s = self.styles.get(self.VECTOR_DEFAULT_STYLE_NAME, None)

        if s is not None:
            # print(f"data_id={data_id}")
            if data_id is not None:
                s = s.set_data_id(data_id)
                return s
            else:
                return s.clone()
        return None

    def style(self, fs, plot_type="map", data_id=None):
        param_info = fs.ds_param_info
        if param_info is not None:
            vd = self.get_param_style(
                param_info,
                scalar=param_info.scalar,
                plot_type=plot_type,
                data_id=data_id,
            )
            # LOG.debug(f"vd={vd}")
            return vd
        return None

    def visdef(self, fs, plot_type="map", data_id=None):
        vd = self.style(fs, plot_type=plot_type, data_id=data_id)
        return vd.to_request() if vd is not None else None

    def style_list(self, fs, plot_type="map"):
        param_info = fs.ds_param_info
        if param_info is not None:
            return self.get_param_style_list(
                param_info, scalar=param_info.scalar, plot_type=plot_type
            )
        return []

    def units_scaler(self, fs, plot_type="map"):
        param_info = fs.ds_param_info
        if param_info is not None:
            p = self._best_param_match(param_info)
            if p is not None and ("all" in p.scaling or plot_type in p.scaling):
                return Scaling.find_item(
                    ParamInfo._grib_get(
                        fs[0], ["units", "shortName"], single_value_as_list=False
                    )
                )

    def _make_defaults(self):
        d = {
            self.SCALAR_DEFAULT_STYLE_NAME: "mcont",
            self.VECTOR_DEFAULT_STYLE_NAME: "mwind",
            self.DIFF_DEFAULT_STYLE_NAME: "mcont",
        }
        for name, verb in d.items():
            if name not in self.styles:
                self.styles[name] = Style(name, Visdef(verb, {}))
        assert self.SCALAR_DEFAULT_STYLE_NAME in self.styles
        assert self.VECTOR_DEFAULT_STYLE_NAME in self.styles
        assert self.DIFF_DEFAULT_STYLE_NAME in self.styles

    def _load(self, param_path, style_path):
        if os.path.exists(style_path):
            with open(style_path, "rt") as f:
                c = yaml.safe_load(f)
                self._load_styles(c)
            if os.path.exists(param_path):
                with open(param_path, "rt") as f:
                    c = yaml.safe_load(f)
                    self._load_params(c, param_path)

    def _load_styles(self, conf):
        for name, d in conf.items():
            vd = []
            # print(f"name={name} d={d}")
            if not isinstance(d, list):
                d = [d]

            # print(f"name={name} d={d}")
            # for mcoast the verb can be missing
            if (
                len(d) == 1
                and isinstance(d[0], dict)
                and (len(d[0]) > 1 or not list(d[0].keys())[0] in PARAM_VISDEF_VERBS)
            ):
                vd.append(Visdef("mcoast", d[0]))
            else:
                for v in d:
                    ((verb, params),) = v.items()
                    vd.append(Visdef(verb, params))
            self.styles[name] = Style(name, vd)

    def _load_params(self, conf, path):
        # TODO: review defaults for maps
        self._make_defaults()

        for d in conf:
            assert isinstance(d, dict)
            # print(f"d={d}")
            p = ParamStyle(d, self)
            for v in [p.style, p.xs_style, p.diff_style]:
                # print(f"v={v}")
                for s in v:
                    if not s in self.styles:
                        raise Exception(
                            f"{self} Invalid style={s} specified in {d}! File={path}"
                        )

            self.params.append(p)

    def is_empty(self):
        return len(self.styles) == 0

    def __str__(self):
        return self.__class__.__name__

    def print(self):
        pass
        # print(f"{self} params=")
        # for k, v in self.params.items():
        #     print(v)
        # print(f"{self} styles=")
        # for k, v in self.styles.items():
        #     print(v)


class GeoView:
    def __init__(self, params, style):
        self.params = copy.deepcopy(params)
        for k in list(self.params.keys()):
            if k.lower() == "coastlines":
                self.params.pop("coastlines", None)
        self.style = style

    def to_request(self):
        v = copy.deepcopy(self.params)
        if self.style is not None and self.style:
            v["coastlines"] = self.style.to_request()
        return mv.geoview(**v)

    def __str__(self):
        t = f"{self.__class__.__name__}[params={self.params}, style={self.style}]"
        return t


class MapConf:
    items = []
    areas = []
    BUILTIN_AREAS = [
        "AFRICA",
        "ANTARCTIC",
        "ARCTIC",
        "AUSTRALASIA",
        "CENTRAL_AMERICA",
        "CENTRAL_EUROPE",
        "EAST_TROPIC",
        "EASTERN_ASIA",
        "EURASIA",
        "EUROPE",
        "GLOBAL",
        "MIDDLE_EAST_AND_INDIA",
        "NORTH_AMERICA",
        "NORTH_ATLANTIC",
        "NORTH_EAST_EUROPE",
        "NORTH_POLE",
        "NORTH_WEST_EUROPE",
        "NORTHERN_AFRICA",
        "PACIFIC",
        "SOUTH_AMERICA",
        "SOUTH_ATLANTIC_AND_INDIAN_OCEAN",
        "SOUTH_EAST_ASIA_AND_INDONESIA",
        "SOUTH_EAST_EUROPE",
        "SOUTH_POLE",
        "SOUTH_WEST_EUROPE",
        "SOUTHERN_AFRICA",
        "SOUTHERN_ASIA",
        "WEST_TROPIC",
        "WESTERN_ASIA",
    ]

    def __init__(self):
        self.areas = {}
        self.style_db = get_db(name="map")

        # load areas
        self._load_areas(os.path.join(ETC_PATH, "areas.yaml"))
        self._load_custom_config()
        self._load_local_config()

    def _load_custom_config(self):
        if CUSTOM_CONF_PATH and CUSTOM_CONF_PATH[-1]:
            self._load_areas(os.path.join(CUSTOM_CONF_PATH[-1], "areas.yaml"))

    def _load_local_config(self):
        if LOCAL_CONF_PATH:
            self._load_areas(os.path.join(LOCAL_CONF_PATH, "areas.yaml"))

    def _load_areas(self, file_path):
        if os.path.exists(file_path):
            with open(file_path, "rt") as f:
                # the file can be empty!
                d = yaml.safe_load(f)
                if isinstance(d, list):
                    for item in d:
                        ((name, conf),) = item.items()
                        self.areas[name] = conf

    def area_names(self):
        r = list(self.areas.keys())
        r.extend([a.lower() for a in self.BUILTIN_AREAS])
        return r

    def find(self, area=None, style=None):
        area_v = "base" if area is None else area
        style_v = "base" if style is None else style
        s = None
        if isinstance(area_v, list):
            if len(area_v) == 4:
                a = {
                    "area_mode": "user",
                    "map_projection": "cylindrical",
                    "map_area_definition": "corners",
                    "area": area_v,
                }
            else:
                raise Exception(
                    "Invalid list specified for area. Required format: [S,W,N,E]"
                )
        else:
            a = self.areas.get(area_v, {})
            if len(a) == 0 and area_v.upper() in self.BUILTIN_AREAS:
                a = {"area_mode": "name", "area_name": area}
        if isinstance(style_v, mv.Request):
            s = style_v
        else:
            s = self.style_db.get_style(style_v)
        return a, s

    def make_geo_view(self, area=None, style=None, plot_type=None):
        if style is None and plot_type == "diff":
            style = "base_diff"

        a, s = self.find(area=area, style=style)
        if a is not None and a:
            a = copy.deepcopy(a)
        else:
            a = {}

        if s is not None:
            if isinstance(s, mv.Request):
                s = Visdef.from_request(s)
            else:
                s = s.clone()
            if plot_type == "stamp":
                s = s.update({"map_grid": "off", "map_label": "off"})
            if s:
                a["coastlines"] = s.to_request()

        return mv.geoview(**a)


class StyleGallery:
    def __init__(self):
        pass

    def to_base64(self, image_path):
        import base64

        with open(image_path, "rb") as img_file:
            return base64.b64encode(img_file.read()).decode("utf-8")

    def build_gallery(self, names, images, row_height="150px"):
        figures = []
        # print(len(names))
        # print(len(images))
        for name, image in zip(names, images):
            src = f"data:image/png;base64,{image}"
            caption = f'<figcaption style="text-align: center; font-size: 0.8em">{name}</figcaption>'
            figures.append(
                f"""
                <figure style="margin: 2px !important;">
                <img src="{src}" style="height: {row_height}">
                {caption}
                </figure>
            """
            )
        return f"""
            <div style="display: flex; flex-flow: row wrap; text-align: center;">
            {''.join(figures)}
            </div>
        """

    def build(self):
        if not is_ipython_active():
            return

        img_size = 120
        img, names = self._build(img_size)

        # reset jupyter output settings
        mv.setoutput("jupyter", **mv.plot.jupyter_args)

        if len(img) > 0:
            from IPython.display import HTML

            return HTML(self.build_gallery(names, img, row_height=f"{img_size}px"))


class MapStyleGallery(StyleGallery):
    def _build(self, img_size):
        img = []
        names = []
        # img_size = 120
        tmp_dir = os.path.join(os.getenv("METVIEW_TMPDIR", ""), "_mapstyle_")
        Path(tmp_dir).mkdir(exist_ok=True)
        for name, d in map_styles().items():
            f_name = os.path.join(tmp_dir, name + ".png")
            if not os.path.exists(f_name):
                view = mv.make_geoview(area=[30, -30, 75, 45], style=name)
                view.update({"MAP_COASTLINE_RESOLUTION": "low"}, sub="coastlines")
                mv.setoutput(
                    mv.png_output(
                        output_name=f_name[:-4],
                        output_width=300,
                        output_name_first_page_number="off",
                    )
                )
                mv.plot(view)
            if os.path.exists(f_name):
                names.append(name)
                img.append(self.to_base64(f_name))

        return (img, names)


class MapAreaGallery(StyleGallery):
    def _build(self, img_size):
        img = []
        names = []
        # img_size = 120
        tmp_dir = os.path.join(os.getenv("METVIEW_TMPDIR", ""), "_maparea_")
        Path(tmp_dir).mkdir(exist_ok=True)
        for name in map_area_names():
            f_name = os.path.join(tmp_dir, name + ".png")
            if not os.path.exists(f_name):
                view = mv.make_geoview(area=name, style="grey_1")
                mv.setoutput(
                    mv.png_output(
                        output_name=f_name[:-4],
                        output_width=300,
                        output_name_first_page_number="off",
                    )
                )
                mv.plot(view)
            if os.path.exists(f_name):
                names.append(name)
                img.append(self.to_base64(f_name))

        return (img, names)


def MAP_CONF():
    global _MAP_CONF
    if _MAP_CONF is None:
        _MAP_CONF = MapConf()
    assert not _MAP_CONF is None
    return _MAP_CONF


def map_styles():
    return MAP_CONF().style_db.styles


def map_style_gallery():
    g = MapStyleGallery()
    return g.build()


def map_area_names():
    return MAP_CONF().area_names()


def map_area_gallery():
    g = MapAreaGallery()
    return g.build()


def make_geoview(**argv):
    return MAP_CONF().make_geo_view(**argv)


def make_eccharts_mcont():
    return mv.mcont(contour_automatic_settings="ecmwf", legend="on")


def get_db(name="param"):
    global _DB
    assert name in _DB
    if _DB[name][0] is None:
        _DB[name][0] = StyleDb(_DB[name][1], _DB[name][2])
    return _DB[name][0]


def find(name):
    db = get_db()
    s = db.styles.get(name, None)
    if s is not None:
        return s.clone()
    else:
        return None


def load_custom_config(conf_dir, force=False):
    global CUSTOM_CONF_PATH
    global _MAP_CONF
    if CUSTOM_CONF_PATH:
        if CUSTOM_CONF_PATH[-1] == conf_dir:
            if force:
                CUSTOM_CONF_PATH.pop()
            else:
                return

    CUSTOM_CONF_PATH.append(conf_dir)
    for k, v in _DB.items():
        if v[0] is not None:
            v[0]._load_custom_config()
    if _MAP_CONF is not None:
        _MAP_CONF = None
        _MAP_CONF = MapConf()


def reset_config():
    global CUSTOM_CONF_PATH
    if CUSTOM_CONF_PATH:
        CUSTOM_CONF_PATH = []
        global _DB
        global _MAP_CONF
        for name, v in _DB.items():
            if v[0] is not None:
                v[0] = None
                get_db(name=name)
        if _MAP_CONF is not None:
            _MAP_CONF = None
            _MAP_CONF = MapConf()


if __name__ == "__main__":
    vd = StyleConf()
    vd.print()
else:
    pass



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from pathlib import Path
import shutil

import pandas as pd
import requests
import yaml

import metview as mv
from metview.metviewpy.indexer import ExperimentIndexer
from metview.metviewpy.indexdb import IndexDb
from metview.track import Track
from metview.metviewpy.param import init_pandas_options
from metview.metviewpy import utils


ETC_PATH = os.path.join(os.path.dirname(__file__), "etc")

# logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")
# logging.basicConfig(level=logging.DEBUG, format="%(levelname)s - %(message)s")
LOG = logging.getLogger(__name__)


class ExperimentDb(IndexDb):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self.fs = {}
        self.vector_loaded = True
        self._indexer = None
        self.fieldset_class = mv.Fieldset
        LOG.debug(f"rootdir_placeholder_value={self.rootdir_placeholder_value}")

    @staticmethod
    def make_from_conf(name, conf, root_dir, db_root_dir, regrid_conf, dataset):
        # LOG.debug(f"conf={conf}")
        db = ExperimentDb(
            name,
            label=conf.get("label", ""),
            desc=conf.get("desc", ""),
            path=conf.get("dir", "").replace(
                IndexDb.ROOTDIR_PLACEHOLDER_TOKEN, root_dir
            ),
            rootdir_placeholder_value=root_dir
            if IndexDb.ROOTDIR_PLACEHOLDER_TOKEN in conf.get("dir", "")
            or "merge" in conf
            else "",
            file_name_pattern=conf.get("fname", ""),
            db_dir=os.path.join(db_root_dir, name),
            merge_conf=conf.get("merge", []),
            mapped_params={v: k for k, v in conf.get("mapped_params", {}).items()},
            regrid_from=regrid_conf.get(name, []),
            blocks={},
            dataset=dataset,
        )
        return db

    def _clone(self):
        return ExperimentDb(
            self.name,
            label=self.label,
            db_dir=self.db_dir,
            mapped_params=self.mapped_params,
            regrid_from=self.regrid_from,
            dataset=self.dataset,
            data_files=self.data_files,
            rootdir_placeholder_value=self.rootdir_placeholder_value,
        )

    @property
    def indexer(self):
        if self._indexer is None:
            self._indexer = ExperimentIndexer(self)
        return self._indexer

    def scan(self, vector=True):
        print(f"Generate index for dataset component={self.name} ...")
        self.data_files = []
        # self.blocks = {}
        self.indexer.scan()

    def load(self, keys=None, vector=True):
        keys = [] if keys is None else keys
        ivk = [x for x in keys if x not in self.indexer.allowed_keys()]
        if ivk:
            raise Exception(
                f"{self} keys={ivk} cannot be used! The allowed set of keys={self.indexer.allowed_keys()}"
            )

        self.load_data_file_list()
        if len(self.data_files) == 0:
            self.scan(vector=True)
            self.load_data_file_list()

        if len(self.blocks) == 0:
            for key in ExperimentIndexer.get_storage_key_list(self.db_dir):
                self.blocks[key] = ExperimentIndexer.read_dataframe(key, self.db_dir)

    def load_data_file_list(self):
        if len(self.data_files) == 0:
            try:
                file_path = os.path.join(self.db_dir, "datafiles.yaml")
                with open(file_path, "rt") as f:
                    self.data_files = yaml.safe_load(f)
                if self.rootdir_placeholder_value:
                    self.data_files = [
                        x.replace(
                            IndexDb.ROOTDIR_PLACEHOLDER_TOKEN,
                            self.rootdir_placeholder_value,
                        )
                        for x in self.data_files
                    ]
                # LOG.debug(f"data_files={self.data_files}")
                for f in self.data_files:
                    assert os.path.isfile(f)
            except:
                pass

    def __getitem__(self, key):
        if isinstance(key, str):
            self.load()
            return self.select_with_name(key)
        return None

    def _filter_blocks(self, dims):
        self.load()
        dfs = {}
        # LOG.debug(f"data_files={self.data_files}")
        # LOG.debug(f"dims={dims}")
        cnt = 0
        for key, df in self.blocks.items():
            # LOG.debug(f"key={key}")
            # df = self._load_block(key)
            # LOG.debug(f"df={df}")
            f_df = self._filter_df(df=df, dims=dims)
            # LOG.debug(f"df={df}"
            if f_df is not None and not f_df.empty:
                cnt += len(f_df)
                # LOG.debug(f" matching rows={len(df)}")
                dfs[key] = f_df

        # LOG.debug(f"total matching rows={cnt}")
        return dfs

    def _extract_fields(self, df, fs, max_count):
        if df.empty:
            return None

        if "_fileIndex3" in df.columns:
            comp_num = 3
        elif "_fileIndex2" in df.columns:
            comp_num = 2
        elif "_fileIndex1" in df.columns:
            comp_num = 1
        else:
            return None

        idx = [[] for k in range(comp_num)]
        comp_lst = list(range(comp_num))
        for row in df.itertuples():
            for comp in comp_lst:
                idx_file = row[-1 - (comp_num - comp - 1) * 2]
                idx_msg = row[-2 - (comp_num - comp - 1) * 2]
                if not idx_file in self.fs:
                    self.fs[idx_file] = mv.read(self.data_files[idx_file])
                fs.append(self.fs[idx_file][idx_msg])
                idx[comp].append(len(fs) - 1)
        # generate a new dataframe
        df = df.copy()
        for k, v in enumerate(idx):
            df[f"_msgIndex{k+1}"] = v
        df.drop([f"_fileIndex{x+1}" for x in range(comp_num)], axis=1, inplace=True)
        return df

    def to_fieldset(self):
        db, fs = self._get_fields({})
        fs._db = db
        return fs

    def get_longname_and_units(self, shortName, paramId):
        return "", ""


class TrackConf:
    def __init__(self, name, conf, data_dir, dataset):
        self.name = name
        self.dataset = dataset
        self.label = self.name
        self.path = conf.get("dir", "").replace(
            IndexDb.ROOTDIR_PLACEHOLDER_TOKEN, data_dir
        )
        self.file_name_pattern = conf.get("fname", "")
        # self.conf_dir = os.path.join("_conf", self.name)
        self.data_files = []
        self.conf = conf

    def load_data_file_list(self):
        if len(self.data_files) == 0:
            self.data_files = utils.get_file_list(
                self.path, file_name_pattern=self.file_name_pattern
            )

    def select(self, name):
        tr = self._make(name)
        if tr is None:
            raise Exception(f"No track is available with name={name}!")
        return tr

    def describe(self):
        self.load_data_file_list()
        init_pandas_options()
        t = {"Name": [], "Suffix": []}
        for f in self.data_files:
            n, s = os.path.splitext(os.path.basename(f))
            t["Name"].append(n)
            t["Suffix"].append(s)
        df = pd.DataFrame.from_dict(t)
        df.set_index("Name", inplace=True)
        return df

    def _make(self, name):
        self.load_data_file_list()
        for f in self.data_files:
            if name == os.path.basename(f).split(".")[0]:
                c = {
                    x: self.conf.get(x, None)
                    for x in [
                        "skiprows",
                        "date_index",
                        "time_index",
                        "lon_index",
                        "lat_index",
                    ]
                }
                return Track(f, **c)
        return None


class Dataset:
    """
    Represents a Dataset
    """

    URL = "https://get.ecmwf.int/repository/test-data/metview/dataset"
    LOCAL_ROOT = os.getenv(
        "MPY_DATASET_ROOT", os.path.join(os.getenv("HOME", ""), "mpy_dataset")
    )
    COMPRESSION = "bz2"

    def __init__(self, name_or_path, load_style=True):
        self.field_conf = {}
        self.track_conf = {}

        self.path = name_or_path
        if any(x in self.path for x in ["/", "\\", "..", "./"]):
            self.name = os.path.basename(self.path)
        else:
            self.name = self.path
            self.path = ""

        assert self.name
        # print(f"name={self.name}")

        # set local path
        if self.path == "":
            self.path = os.path.join(self.LOCAL_ROOT, self.name)

        # If the path does not exists, it must be a built-in dataset. Data will be
        # downloaded into path.
        if not os.path.isdir(self.path):
            if self.check_remote():
                self.fetch(forced=True)
            else:
                raise Exception(
                    f"Could not find dataset={self.name} either under path={self.path} or on data server"
                )
        # WARN: we do not store the dataset in the CACHE any more. Check code versions before
        # 09082021 to see how the CACHE was used.

        if load_style:
            self.load_style()

        self.load()

        for _, c in self.field_conf.items():
            LOG.debug(f"{c}")

    @staticmethod
    def load_dataset(*args, **kwargs):
        return Dataset(*args, **kwargs)

    def check_remote(self):
        try:
            return (
                requests.head(
                    f"{self.URL}/{self.name}/conf.tar", allow_redirects=True
                ).status_code
                == 200
            )
        except:
            return False

    def load(self):
        data_dir = os.path.join(self.path, "data")
        index_dir = os.path.join(self.path, "index")
        data_conf_file = os.path.join(self.path, "data.yaml")
        with open(data_conf_file, "rt") as f:
            d = yaml.safe_load(f)
            regrid_conf = d.get("regrid", {})
            for item in d["experiments"]:
                ((name, conf),) = item.items()
                if conf.get("type") == "track":
                    self.track_conf[name] = TrackConf(name, conf, data_dir, self)
                else:
                    c = ExperimentDb.make_from_conf(
                        name, conf, data_dir, index_dir, regrid_conf, self
                    )
                    self.field_conf[c.name] = c

    def scan(self, name=None):
        # indexer = ExperimentIndexer()
        if name:
            if name in self.field_conf:
                self.field_conf[name].scan()
                # indexer.scan(self.field_conf[name], to_disk=True)
        else:
            for name, c in self.field_conf.items():
                LOG.info("-" * 40)
                c.scan()
                # indexer.scan(c, to_disk=True)

    def find(self, name, comp="field"):
        if comp == "all":
            f = self.field_conf.get(name, None)
            if f is not None:
                return f
            else:
                return self.track_conf.get(name, None)
        elif comp == "field":
            return self.field_conf.get(name, None)
        elif comp == "track":
            return self.track_conf.get(name, None)
        else:
            return None

    def describe(self):
        init_pandas_options()
        print("Dataset components:")
        t = {"Component": [], "Description": []}
        for _, f in self.field_conf.items():
            t["Component"].append(f.name)
            t["Description"].append(f.desc)
        for _, f in self.track_conf.items():
            t["Component"].append(f.name)
            t["Description"].append("Storm track data")
        df = pd.DataFrame.from_dict(t)
        df.set_index("Component", inplace=True)
        return df

    def fetch(self, forced=False):
        if not os.path.isdir(self.path):
            Path(self.path).mkdir(0o755, parents=True, exist_ok=True)

        files = {
            "conf.tar": ["data.yaml", "conf"],
            f"index.tar.{self.COMPRESSION}": ["index"],
            f"data.tar.{self.COMPRESSION}": ["data"],
        }

        checked = False
        for src, targets in files.items():
            # if forced or not utils.CACHE.all_exists(targets, self.path):
            if forced:
                if not checked and not self.check_remote():
                    raise Exception(
                        f"Could not find dataset={self.name} on data server"
                    )
                else:
                    checked = True

                remote_file = os.path.join(self.URL, self.name, src)
                target_file = os.path.join(self.path, src)
                # LOG.debug(f"target_file={target_file}")
                try:
                    # print("Download data ...")
                    utils.download(remote_file, target_file)
                    print(f"Unpack data ... {src}")
                    utils.unpack(target_file, remove=True)
                    # TODO: we skip the reference creation to make things faster. Enable it when
                    # it is needed!
                    # utils.CACHE.make_reference(targets, self.path)
                except:
                    # if os.exists(target_file):
                    #     os.remove(target_file)
                    raise Exception(f"Failed to download file={remote_file}")

    def __getitem__(self, key):
        if isinstance(key, str):
            r = self.find(key, comp="all")
            if r is None:
                raise Exception(f"No component={key} found in {self}")
            return r
        return None

    def load_style(self):
        conf_dir = os.path.join(self.path, "conf")
        mv.style.load_custom_config(conf_dir, force=True)

    def __str__(self):
        return f"{self.__class__.__name__}[name={self.name}]"


def create_dataset_template(name_or_path):
    path = name_or_path
    if any(x in path for x in ["/", "\\", "..", "./"]):
        name = os.path.basename(path)
    else:
        name = path
        path = ""

    if path == "":
        path = os.path.join(Dataset.LOCAL_ROOT, name)

    if not os.path.exists(path):
        os.mkdir(path)
    else:
        if not os.path.isdir(path):
            raise Exception(f"path must be a directory!")
        if os.path.exists(os.path.join("path", "data.yaml")):
            raise Exception(
                f"The specified dataset directory={path} already exists and is not empty!"
            )

    # create dirs
    for dir_name in ["conf", "data", "index"]:
        os.mkdir(os.path.join(path, dir_name))

    # copy files
    files = {
        "params.yaml": ("conf", ""),
        "param_styles.yaml": ("conf", ""),
        "areas.yaml": ("conf", ""),
        "map_styles_template.yaml": ("conf", "map_styles.yaml"),
        "dataset_template.yaml": ("", "data.yaml"),
    }
    for src_name, target in files.items():
        target_dir = os.path.join(path, target[0]) if target[0] else path
        target_name = target[1] if target[1] else src_name
        shutil.copyfile(
            os.path.join(ETC_PATH, src_name), os.path.join(target_dir, target_name)
        )



#
# (C) Copyright 2020- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import math

import metview as mv

LOG = logging.getLogger(__name__)


class Layout:

    GRID_DEF = {
        1: [1, 1],
        2: [2, 1],
        3: [3, 1],
        4: [2, 2],
        8: [3, 3],
        9: [3, 3],
        10: [4, 3],
    }

    def _grid_row_col(self, page_num=0, rows=None, columns=None, layout=None):
        if rows is None and columns is None:
            if layout is not None and layout != "":
                v = layout.split("x")
                if len(v) == 2:
                    try:
                        r = int(v[0])
                        c = int(v[1])
                        # LOG.debug(f"r{r} c={c}")
                    except:
                        raise Exception(f"Invalid layout specification ({layout}")
                    if page_num > 0 and r * c < page_num:
                        raise Exception(
                            f"Layout specification={layout} does not match number of scenes={page_num}"
                        )
            else:
                if page_num in self.GRID_DEF:
                    r, c = self.GRID_DEF[page_num]
                elif page_num >= 1:
                    r, c = self._compute_row_col(page_num)
                else:
                    raise Exception(f"Cannot create a layout for {page_num} pages!")
        return r, c

    def build_grid(self, page_num=0, rows=None, columns=None, layout=None, view=None):
        r, c = self._grid_row_col(
            page_num=page_num, rows=rows, columns=columns, layout=layout
        )

        if isinstance(view, mv.style.GeoView):
            v = view.to_request()
        else:
            v = view

        return self._build_grid(rows=r, columns=c, view=v)

    def _build_grid(self, rows=1, columns=1, view=None):
        assert rows >= 1 and columns >= 1
        if rows == 1 and columns == 1:
            return mv.plot_superpage(pages=[mv.plot_page(view=view)])
        else:
            return mv.plot_superpage(
                pages=mv.mvl_regular_layout(
                    view, columns, rows, 1, 1, [5, 100, 15, 100]
                )
            )

    def _compute_row_col(self, page_num):
        # has been checked for 100
        r = int(math.sqrt(page_num))
        c = r
        if c * r < page_num:
            if (page_num - c * r) % c == 0:
                c += (page_num - c * r) // c
            else:
                c += 1 + (page_num - c * r) // c
        return (r, c)

    def build_diff(self, view):
        page_1 = mv.plot_page(top=5, bottom=50, left=25, right=75, view=view)

        page_2 = mv.plot_page(top=52, bottom=97, right=50, view=view)

        page_3 = mv.plot_page(top=52, bottom=97, left=50, right=100, view=view)

        return mv.plot_superpage(pages=[page_1, page_2, page_3])

    def build_xs(self, line, map_view):
        xs_view = mv.mxsectview(
            line=line,
            bottom_level=1000,
            top_level=150
            # wind_perpendicular : "off",
            # wind_parallel :"on",
            # wind_intensity :"off"
        )

        page_1 = mv.plot_page(top=35 if map_view is not None else 5, view=xs_view)

        if map_view is not None:
            page = mv.plot_page(top=5, bottom=35, view=map_view)
            return mv.plot_superpage(pages=[page_1, page])
        else:
            return mv.plot_superpage(pages=[page_1])

    def build_xs_avg(
        self,
        area,
        direction,
        bottom_level,
        top_level,
        vertical_scaling,
        axis_label_height=0.4,
    ):
        axis = mv.maxis(axis_tick_label_height=axis_label_height)

        return mv.maverageview(
            area=area,
            direction=direction,
            bottom_level=bottom_level,
            top_level=top_level,
            vertical_scaling=vertical_scaling,
            horizontal_axis=axis,
            vertical_axis=axis,
        )

    def build_stamp(self, page_num=0, layout="", view=None):

        if True:
            coast_empty = mv.mcoast(
                map_coastline="off", map_grid="off", map_label="off"
            )

            empty_view = mv.geoview(
                page_frame="off",
                subpage_frame="off",
                coastlines=coast_empty,
                subpage_x_position=40,
                subpage_x_length=10,
                subpage_y_length=10,
            )

            title_page = mv.plot_page(
                top=0, bottom=5, left=30, right=70, view=empty_view
            )

        r, c = self._grid_row_col(page_num=page_num, layout=layout)

        pages = mv.mvl_regular_layout(view, c, r, 1, 1, [5, 100, 15, 100])
        pages.append(title_page)
        return mv.plot_superpage(pages=pages)

        # g = self.build_grid(page_num=page_num, layout=layout, view=view)
        # return g

    def build_rmse(self, xmin, xmax, ymin, ymax, xtick, ytick, xtitle, ytitle):

        horizontal_axis = mv.maxis(
            axis_type="date",
            axis_orientation="horizontal",
            axis_position="bottom",
            # axis_title               :  'on',
            # axis_title_height        :   0.4,
            axis_grid="on",
            axis_grid_colour="grey",
            # axis_title_text          :  xtitle,
            # axis_title_quality       :  'high',
            axis_tick_interval=xtick,
            axis_tick_label_height=0.6,
            axis_date_type="days",
            axis_years_label="off",
            axis_months_label="off",
            axis_days_label_height="0.3",
        )

        vertical_axis = mv.maxis(
            axis_orientation="vertical",
            axis_position="left",
            axis_grid="on",
            axis_grid_colour="grey",
            axis_title="on",
            axis_title_height=0.4,
            axis_title_text=ytitle,
            # axis_title_quality       :  'high',
            axis_tick_interval=ytick,
            axis_tick_label_height=0.3,
        )

        cview = mv.cartesianview(
            page_frame="off",
            x_axis_type="date",
            y_axis_type="regular",
            y_automatic="off",
            x_automatic="off",
            y_min=ymin,
            y_max=ymax,
            x_date_min=xmin,
            x_date_max=xmax,
            horizontal_axis=horizontal_axis,
            vertical_axis=vertical_axis,
        )

        return cview

    def build_xy(self, xmin, xmax, ymin, ymax, xtick, ytick, xtitle, ytitle):

        horizontal_axis = mv.maxis(
            axis_orientation="horizontal",
            axis_position="bottom",
            axis_title="on",
            axis_title_height=0.5,
            axis_title_text=xtitle,
            # axis_title               :  'on',
            # axis_title_height        :   0.4,
            axis_grid="on",
            axis_grid_colour="grey",
            # axis_title_text          :  xtitle,
            # axis_title_quality       :  'high',
            axis_tick_interval=xtick,
            axis_tick_label_height=0.6,
            axis_date_type="days",
            axis_years_label="off",
            axis_months_label="off",
            axis_days_label_height="0.3",
        )

        vertical_axis = mv.maxis(
            axis_orientation="vertical",
            axis_position="left",
            axis_grid="on",
            axis_grid_colour="grey",
            axis_title="on",
            axis_title_height=0.6,
            axis_title_text=ytitle,
            # axis_title_quality       :  'high',
            axis_tick_interval=ytick,
            axis_tick_label_height=0.3,
        )

        cview = mv.cartesianview(
            page_frame="off",
            x_axis_type="regular",
            y_axis_type="regular",
            y_automatic="off",
            x_automatic="off",
            y_min=ymin,
            y_max=ymax,
            x_min=xmin,
            x_max=xmax,
            horizontal_axis=horizontal_axis,
            vertical_axis=vertical_axis,
        )

        return cview

    @staticmethod
    def compute_axis_range(v_min, v_max):
        count = 15
        d = math.fabs(v_max - v_min)
        if d > 0:
            b = d / count
            n = math.floor(math.log10(b))
            v = b / math.pow(10, n)
            # print("d={} b={} n={} v={}".format(d,b,n,v))

            if v <= 1:
                v = 1
            elif v <= 2:
                v = 2
            elif v <= 5:
                v = 5
            else:
                v = 10

            bin_size = v * math.pow(10, n)
            bin_start = math.ceil(v_min / bin_size) * bin_size
            if bin_start >= v_min and math.fabs(bin_start - v_min) < bin_size / 10000:
                bin_start = bin_start - bin_size / 100000
                av = v_min
            else:
                bin_start = bin_start - bin_size
                av = bin_start

            max_iter = 100
            act_iter = 0
            while av < v_max:
                av += bin_size
                act_iter += 1
                if act_iter > max_iter:
                    return (0, v_min, v_max)

            bin_end = av  # + bin_size / 100000
            return bin_size, bin_start, bin_end
        else:
            return (0, v_min, v_max)



#
# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import metview as mv

LOG = logging.getLogger(__name__)

FC_TIME_PART = "Step: +<grib_info key='step' />h Valid: <grib_info key='valid-date' format='%Y-%m-%d %H' />"
LEVEL_PART = "Lev: <grib_info key='level' /> (<grib_info key='levelType' />)"


class Title:
    def __init__(self, font_size=0.4):
        self.font_size = font_size

    def _make_font_size(self, font_size):
        if font_size is None:
            return self.font_size
        else:
            return font_size

    def build(self, data, font_size=None):
        font_size = self._make_font_size(font_size)
        label = ""
        if data:
            if not isinstance(data, list):
                data = [data]

            lines = []
            lines = {"text_line_count": len(data)}
            for i, d_item in enumerate(data):
                # print(f"d_item={d_item}")
                if isinstance(d_item, tuple):
                    d = d_item[0]
                    data_id = d_item[1]
                else:
                    d = d_item
                    data_id = None

                param = d.ds_param_info
                if param is not None:
                    if param.meta.get("typeOfLevel", "") == "surface":
                        # lines.append(self.build_surface_fc(d.experiment.label, d.param.name, condition=cond))
                        lines[f"text_line_{i+1}"] = self.build_surface_fc(
                            d.label, param.name, data_id=data_id
                        )
                    else:
                        # lines.append(self.build_upper_fc(d.experiment.label, d.param.name, condition=cond))
                        lines[f"text_line_{i+1}"] = self.build_upper_fc(
                            d.label, param.name, data_id=data_id
                        )

            # print(f"line={lines}")
            # return mv.mtext(text_lines=lines, text_font_size=font_size)
            return mv.mtext(**lines, text_font_size=font_size)

        return mv.mtext(
            {
                "text_line_1": f"""{label} Par: <grib_info key='shortName'/> Lev: <grib_info key='level'/> (<grib_info key='levelType'/>) Step: +<grib_info key='step'/>h Valid: <grib_info key='valid-date' format='%Y-%m-%d %H'/>""",
                "text_font_size": font_size,
            }
        )

    def _build_condition_str(self, condition):
        if condition:
            t = "where="
            for k, v in condition.items():
                t += f"'{k}={v}'"
            return t
        return str()

    def _add_grib_info(self, t, data_id):
        if data_id is not None and data_id != "":
            t = t.replace("<grib_info", f"<grib_info id='{data_id}'")
        return t

    def build_surface_fc(self, label, par, condition=None, data_id=None):
        t = FC_TIME_PART
        t = self._add_grib_info(t, data_id)
        return f"""{label} Par: {par} {t}"""

    def build_upper_fc(self, label, par, condition=None, data_id=None):
        lev = LEVEL_PART
        t = FC_TIME_PART
        lev = self._add_grib_info(lev, data_id)
        t = self._add_grib_info(t, data_id)
        return f"""{label} Par: {par} {lev} {t}"""

    def build_xs(self, data):
        if data:
            if not isinstance(data, list):
                data = [data]

            lines = []
            for d in data:
                line = d.label
                # print(f"label={d.label}")
                param = d.ds_param_info
                if param is not None:
                    # print(f"meta={param.meta}")
                    line += f" Par: {param.name}"
                    meta = param.meta
                    if meta.get("mars.type", "") == "an":
                        pass
                    else:
                        r = meta.get("date", 0)
                        h = meta.get("time", 0)
                        s = meta.get("step", "")
                        line += f" Run: {r} {h} UTC Step: +{s}h"
                lines.append(line)

            return mv.mtext(text_lines=lines, text_font_size=self.font_size)

        return mv.mtext(text_font_size=self.font_size)

    def build_stamp(self, data, member=""):
        if data:
            if not isinstance(data, list):
                data = [data]

            # exp_conf = data[0].experiment
            label = data[0].label

            member_txt = ""
            if member:
                if member == "0":
                    member_txt = "CF"
                else:
                    member_txt = f"PF {member}"

            return mv.mtext(
                {
                    "text_line_1": f"""{label} {member_txt} Step: +<grib_info key='step'/>h Valid: <grib_info key='valid-date' format='%Y-%m-%d %H'/>""",
                    "text_font_size": self.font_size,
                }
            )

        return mv.mtext(
            {
                "text_font_size": self.font_size,
            }
        )

    def build_rmse(self, ref, data):
        if data:
            if not isinstance(data, list):
                data = [data]

            lines = []
            for d in data:
                line = "RMSE"
                if ref.label:
                    line += f"(ref={ref.label})"
                if d.label:
                    line += f" {d.label}"

                # print(f"label={d.label}")
                param = d.ds_param_info
                if param is not None:
                    # print(f"meta={param.meta}")
                    line += f" Par: {param.name}"
                    meta = param.meta
                    if meta.get("mars.type", "") == "an":
                        pass
                    else:
                        r = meta.get("date", 0)
                        h = meta.get("time", 0)
                        line += f" Run: {r} {h} UTC"
                lines.append(line)

            return mv.mtext(text_lines=lines, text_font_size=self.font_size)

        return mv.mtext(text_font_size=self.font_size)

    def build_cdf(self, data):
        if data:
            if not isinstance(data, list):
                data = [data]
            lines = []
            for d in data:
                line = f"CDF {d.label}"
                # print(f"label={d.label}")
                param = d.ds_param_info
                if param is not None:
                    # print(f"meta={param.meta}")
                    line += f" Par: {param.name}"
                    meta = param.meta
                    if meta.get("mars.type", "") == "an":
                        pass
                    else:
                        r = meta.get("date", 0)
                        h = meta.get("time", 0)
                        line += f" Run: {r} {h} UTC"
                lines.append(line)

            return mv.mtext(text_lines=lines, text_font_size=self.font_size)

        return mv.mtext(text_font_size=self.font_size)



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os


# if we're running pytest, it will need the fieldset functionality, so detect if it
# is running and if so, import the user-facing functions


def running_from_pytest():
    from inspect import stack

    call_stack = [s.function for s in stack()]
    return "pytest_collection" in call_stack


if "METVIEW_PYTHON_ONLY" in os.environ or running_from_pytest():

    from . import fieldset

    fieldset.bind_functions(globals(), module_name=__name__)



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

from metview.metviewpy import utils


def test_utils_date():

    d = utils.date_from_str("20210204")
    assert d == datetime.datetime(2021, 2, 4)

    d = utils.date_from_str("2021-02-04")
    assert d == datetime.datetime(2021, 2, 4)

    d = utils.date_from_str("2021-02-04 06")
    assert d == datetime.datetime(2021, 2, 4, 6, 0, 0)

    d = utils.date_from_str("2021-02-04 06:14")
    assert d == datetime.datetime(2021, 2, 4, 6, 14, 0)

    d = utils.date_from_str("2021-02-04 06:14:32")
    assert d == datetime.datetime(2021, 2, 4, 6, 14, 32)

    d = utils.date_from_str("20210204.25")
    assert d == datetime.datetime(2021, 2, 4, 6, 0, 0)

    d = utils.date_from_str("402")
    assert d == (4, 2)

    d = utils.date_from_str("0402")
    assert d == (4, 2)

    try:
        d = utils.date_from_str("0432")
        assert False
    except:
        pass

    d = utils.date_from_str("apr-02")
    assert d == (4, 2)

    d = utils.date_from_str("Apr-02")
    assert d == (4, 2)

    try:
        d = utils.date_from_str("afr-02")
        assert False
    except:
        pass


def test_utils_time():

    d = utils.time_from_str("6")
    assert d == datetime.time(6, 0, 0)

    d = utils.time_from_str("06")
    assert d == datetime.time(6, 0, 0)

    d = utils.time_from_str("14")
    assert d == datetime.time(14, 0, 0)

    d = utils.time_from_str("600")
    assert d == datetime.time(6, 0, 0)

    d = utils.time_from_str("612")
    assert d == datetime.time(6, 12, 0)

    d = utils.time_from_str("1100")
    assert d == datetime.time(11, 0, 0)

    d = utils.time_from_str("1457")
    assert d == datetime.time(14, 57, 0)

    d = utils.time_from_str("6:12")
    assert d == datetime.time(6, 12, 0)

    d = utils.time_from_str("06:12")
    assert d == datetime.time(6, 12, 0)

    d = utils.time_from_str("14:57")
    assert d == datetime.time(14, 57, 0)


def test_has_globbing():

    assert utils.has_globbing("a*") == True
    assert utils.has_globbing("a?") == True
    assert utils.has_globbing("my_path/a*.grib") == True
    assert utils.has_globbing("my_path/[Aa].grib") == True
    assert utils.has_globbing("my_path/[a-m].grib") == True
    assert utils.has_globbing("my_path/test.grib") == False
    assert utils.has_globbing("my_path/te[st.grib") == False
    assert utils.has_globbing("my_path/tes]t.grib") == False
    # assert utils.has_globbing("my_path/t]e[st.grib") == False



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np

import metview as mv


def concat(*args):
    if len(args) <= 1:
        raise ValueError(f"concat requires at least 2 arguments, got only {len(args)}")

    vals = list(args)
    if vals[0] is None:
        vals.pop(0)
        if len(vals) == 1:
            return vals[0]

    # concatenate strings
    if isinstance(vals[0], str):
        return "".join([str(v) for v in vals])
    elif isinstance(vals[0], np.ndarray):
        return np.concatenate(vals)
    elif isinstance(
        vals[0],
        (mv.Fieldset, mv.bindings.Geopoints, mv.bindings.GeopointSet, mv.bindings.Bufr),
    ):
        return mv.merge(*vals)
    elif isinstance(vals[0], list):
        first = []
        for v in vals:
            if isinstance(v, list):
                first.extend(v)
            else:
                first.append(v)
        return first

    raise ValueError(f"concat failed to handle the specified arguments={args}")


def index4(vals, start, stop, step, num):
    """
    Return a boolean index ndarray to select a subset of elements from ``vals``.

    Parameters
    ----------
    vals: ndarray
        Input array.
    start: int
        Start index.
    stop: int
        Stop index.
    step: int
        Step.
    num: int
        Number of elements to be selected for each step.

    Returns
    -------
    ndarray
        Boolean index array

    Examples
    --------

    >>> import numpy as np
    >>> import metview
    >>> vals = np.array(list(range(12)))
    >>> r = index4(vals, 0, 11, 4, 2)
    [ True  True  True  True  True  True False False False False False False]
    a[]
    >>> vals[r]
    array([0, 1, 4, 5, 8, 9])

    """
    num = min(num, step)
    m = np.zeros(len(vals), dtype=bool)
    for i in range(start, stop, step):
        m[i : i + num] = 1
    return m



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import re
import pandas as pd

from .indexer import GribIndexer
from .ipython import is_ipython_active
from .utils import is_fieldset_type

# logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")
# logging.basicConfig(level=logging.DEBUG, format="%(levelname)s - %(message)s")
LOG = logging.getLogger(__name__)

PANDAS_ORI_OPTIONS = {}


def init_pandas_options():
    global PANDAS_ORI_OPTIONS
    if len(PANDAS_ORI_OPTIONS) == 0:
        opt = {
            "display.max_colwidth": 300,
            "display.colheader_justify": "center",
            "display.max_columns": 100,
            "display.max_rows": 500,
            "display.width": None,
        }
        for k, _ in opt.items():
            PANDAS_ORI_OPTIONS[k] = pd.get_option(k)
        for k, v in opt.items():
            pd.set_option(k, v)


def reset_pandas_options():
    global PANDAS_ORI_OPTIONS
    if len(PANDAS_ORI_OPTIONS) > 0:
        for k, v in PANDAS_ORI_OPTIONS.items():
            pd.set_option(k, v)
        PANDAS_ORI_OPTIONS = {}


class ParamInfo:
    SUFFIXES = {
        "hPa": "isobaricInhPa",
        "hpa": "isobaricInhPa",
        "K": "theta",
        "ml": "hybrid",
    }
    LEVEL_TYPES = {"sfc": "surface", "pl": "isobaricInhPa", "ml": "hybrid"}
    LEVEL_RE = re.compile(r"(\d+)")
    NUM_RE = re.compile(r"[0-9]+")
    SURF_RE = re.compile(r"^\d+\w+")
    # SURF_NAME_MAPPER = {"t2": "2t", "q2": "2q", "u10": "10u", "v10": "10v"}
    KNOWN_SURF_NAMES = [
        "2t",
        "2q",
        "10u",
        "10v",
        "100u",
        "100v",
        "200u",
        "200v",
        "msl",
        "wind10m",
        "wind100m",
        "wind200m",
    ]
    VECTOR_NAMES = [
        "wind100m",
        "wind200m",
        "wind10m",
        "wind3d",
        "wind",
    ]  # the longest ones first

    def __init__(self, name, meta=None, scalar=None):
        self.name = name
        self.scalar = scalar if scalar is not None else True
        self.meta = {} if meta is None else meta
        if len(self.meta) == 0:
            self.meta["shortName"] = name

    def make_filter(self):
        dims = {}
        if self.name:
            dims["shortName"] = [self.name]
        for n in ["level", "typeOfLevel"]:
            v = self.meta.get(n, None)
            if v is not None:
                dims[n] = [v]
        return dims

    @staticmethod
    def build_from_name(full_name, param_level_types=None):
        full_name = full_name
        name = full_name
        level = None
        level_type = ""

        # the name is a known param name
        if param_level_types:
            if name in param_level_types:
                lev_t = param_level_types.get(name, [])
                meta = {}
                if len(lev_t) == 1:
                    meta = {"typeOfLevel": lev_t[0], "level": None}
                scalar = not name in ParamInfo.VECTOR_NAMES
                return ParamInfo(name, meta=meta, scalar=scalar)

        t = full_name
        # surface fields
        if t in ParamInfo.KNOWN_SURF_NAMES or ParamInfo.SURF_RE.match(t) is not None:
            level_type = "surface"

        else:
            # guess the level type from the suffix
            for k, v in ParamInfo.SUFFIXES.items():
                if full_name.endswith(k):
                    level_type = v
                    t = full_name[: -(len(k))]
                    break

            # recognise vector params
            for v in ParamInfo.VECTOR_NAMES:
                if t.startswith(v):
                    name = v
                    t = t[len(v) :]
                    break

            # determine level value
            m = ParamInfo.LEVEL_RE.search(t)
            if m and m.groups() and len(m.groups()) == 1:
                level = int(m.group(1))
                if level_type == "" and level > 10:
                    level_type = "isobaricInhPa"
                if name == full_name:
                    name = ParamInfo.NUM_RE.sub("", t)

        # check param name in the conf
        if param_level_types:
            if not name in param_level_types:
                raise Exception(
                    f"Param={name} (guessed from name={full_name}) is not found in dataset!"
                )

            lev_t = param_level_types.get(name, [])
            if lev_t:
                if not level_type and len(lev_t) == 1:
                    level_type = lev_t[0]
                elif level_type and level_type not in lev_t:
                    raise Exception(
                        f"Level type cannot be guessed from param name={full_name}!"
                    )

        if level_type == "":
            level = None
        scalar = not name in ParamInfo.VECTOR_NAMES

        LOG.debug(f"scalar={scalar}")
        meta = {"level": level, "typeOfLevel": level_type}
        return ParamInfo(name, meta=meta, scalar=scalar)

    @staticmethod
    def build_from_fieldset(fs):
        assert is_fieldset_type(fs)
        f = fs[0:3] if len(fs) >= 3 else fs
        m = ParamInfo._grib_get(f, GribIndexer.DEFAULT_ECC_KEYS)
        name = level = lev_type = ""
        scalar = True

        meta_same = True
        for x in m.keys():
            if x != "shortName" and m[x].count(m[x][0]) != len(m[x]):
                same = False
                break
        if meta_same:
            if len(m["shortName"]) == 3 and m["shortName"] == ["u", "v", "w"]:
                name = "wind3d"
                scalar = False
            elif len(m["shortName"]) >= 2:
                if m["shortName"][0:2] == ["u", "v"]:
                    name = "wind"
                    scalar = False
                elif m["shortName"][0:2] == ["10u", "10v"]:
                    name = "wind10m"
                    m["level"][0] = 0
                    m["typeOfLevel"][0] = "sfc"
                    scalar = False
        if not name:
            name = m["shortName"][0]

        if name:
            return ParamInfo(name, meta={k: v[0] for k, v in m.items()}, scalar=scalar)
        else:
            return None

    def _meta_match(self, meta, key):
        local_key = key if key != "levelist" else "level"
        if (
            key in meta
            and meta[key] is not None
            and meta[key]
            and local_key in self.meta
        ):
            # print(f"local={self.meta[local_key]} other={meta[key]}")
            if isinstance(meta[key], list):
                return str(self.meta[local_key]) in meta[key]
            else:
                return meta[key] == str(self.meta[local_key])
        else:
            return False

    def match(self, name, meta):
        # print(f"{self}, name={name}, meta={meta}")
        r = 0
        if self.name == name:
            r += 3
        for n in ["shortName", "paramId"]:
            if self._meta_match(meta, n):
                r += 1
        # we only check the rest if the param is ok
        if r > 0:
            if self._meta_match(meta, "typeOfLevel"):
                r += 1
                if self._meta_match(meta, "levelist"):
                    r += 1

        return r

    def update_meta(self, meta):
        self.meta = {**meta, **self.meta}

    @staticmethod
    def _grib_get(f, keys, single_value_as_list=True):
        md = f.grib_get(keys, "key")
        m = {}
        for k, v in zip(keys, md):
            key_val = k.split(":")[0]
            val = v
            if k.endswith(":l"):
                val = []
                for x in v:
                    try:
                        val.append(int(x))
                    except:
                        val.append(None)
            if not single_value_as_list and len(val) == 1:
                val = val[0]
            m[key_val] = val
        return m

    def __str__(self):
        return "{}[name={}, scalar={}, meta={}]".format(
            self.__class__.__name__, self.name, self.scalar, self.meta
        )


class ParamDesc:
    def __init__(self, name):
        self.db = None
        # self.name = name
        self.md = {}
        self.levels = {}
        self._short_name = None
        self._param_id = None
        self._long_name = None
        self._units = None

    def load(self, db):
        raise NotImplementedError

    def _parse(self, md):
        if "level" in md and len(md["level"]) > 0:
            df = pd.DataFrame(md)
            md.pop("typeOfLevel")
            md.pop("level")

            for md_key in list(md.keys()):
                d = list(df[md_key].unique())
                self.md[md_key] = d

            lev_types = list(df["typeOfLevel"].unique())
            for t in lev_types:
                # print(f" t={t}")
                self.levels[t] = []
                q = f"typeOfLevel == '{t}'"
                # print(q)
                dft = df.query(q)
                if dft is not None:
                    self.levels[t] = list(dft["level"].unique())

            for k, v in self.md.items():
                self.md[k] = sorted(v)

            for k, v in self.levels.items():
                self.levels[k] = sorted(v)

    @property
    def short_name(self):
        if self._short_name is None:
            self._short_name = ""
            if self.md["shortName"]:
                self._short_name = self.md["shortName"][0]
        return self._short_name

    @property
    def param_id(self):
        if self._param_id is None:
            self._param_id = ""
            if self.md["paramId"]:
                self._param_id = self.md["paramId"][0]
        return self._param_id

    @property
    def long_name(self):
        if self._long_name is None:
            self._long_name = ""
            if self.db is not None:
                self._long_name, self._units = self.db.get_longname_and_units(
                    self.short_name, self.param_id
                )
        return self._long_name

    @property
    def units(self):
        if self._units is None:
            self._units = ""
            if self.db:
                self._long_name, self._units = self.db.get_longname_and_units(
                    self.short_name, self.param_id
                )
        return self._units

    @staticmethod
    def describe(db, param=None, no_print=False):
        labels = {"marsClass": "class", "marsStream": "stream", "marsType": "type"}
        in_jupyter = is_ipython_active()

        # describe all the params
        if param is None:
            t = {"parameter": [], "typeOfLevel": [], "level": []}
            need_number = False
            for k, v in db.param_meta.items():
                if not v.md.get("number", None) in [["0"], [None]]:
                    need_number = True
                    break

            for k, v in db.param_meta.items():
                t["parameter"].append(k)
                if len(v.levels) > 1:
                    lev_type = ""
                    level = ""
                    cnt = 0
                    for md_k, md in v.levels.items():
                        if in_jupyter:
                            lev_type += md_k + "<br>"
                            level += str(ParamDesc.format_list(md)) + "<br>"
                        else:
                            prefix = " " if cnt > 0 else ""
                            lev_type += prefix + f"[{cnt+1}]:" + md_k
                            level += (
                                prefix + f"[{cnt+1}]:" + str(ParamDesc.format_list(md))
                            )
                        cnt += 1
                    t["typeOfLevel"].append(lev_type)
                    t["level"].append(level)
                else:
                    for md_k, md in v.levels.items():
                        t["typeOfLevel"].append(md_k)
                        t["level"].append(ParamDesc.format_list(md))

                for md_k, md in v.md.items():
                    if md_k != "number" or need_number:
                        md_k = labels.get(md_k, md_k)
                        if not md_k in t:
                            t[md_k] = []
                        t[md_k].append(ParamDesc.format_list(md))

            if in_jupyter:
                txt = ParamDesc._make_html_table(t)
                from IPython.display import HTML

                return HTML(txt)
            else:
                df = pd.DataFrame.from_dict(t)
                df = df.set_index(["parameter"])
                init_pandas_options()
                if not no_print:
                    print(df)
                return df

        # specific param
        else:
            v = None
            if isinstance(param, str):
                v = db.param_meta.get(param, None)
            elif isinstance(param, int):
                v = db.param_id_meta(param)

            if v is None or len(v.md) == 0:
                print(f"No shortName/paramId={param} found in data!")
                return

            #  if v is not None:
            t = {
                "key": ["shortName"],
                "val": [v.short_name],
            }

            if v.long_name != "" or v.units != "":
                t["key"].append("name")
                t["val"].append(v.long_name)

            t["key"].append("paramId")
            t["val"].append(v.param_id)
            # ParamDesc.format_list(v.md["shortName"], full=True),

            if v.long_name != "" or v.units != "":
                t["key"].append("units")
                t["val"].append(v.units)

            add_cnt = len(v.levels) > 1
            cnt = 0
            for md_k, md in v.levels.items():
                t["key"].append("typeOfLevel" + (f"[{cnt+1}]" if add_cnt else ""))
                t["val"].append(md_k)
                t["key"].append("level" + (f"[{cnt+1}]" if add_cnt else ""))
                t["val"].append(ParamDesc.format_list(md, full=True))
                cnt += 1

            for kk, md_v in v.md.items():
                if kk == "number" and md_v == ["0"]:
                    continue
                if not kk in ["shortName", "paramId"]:
                    t["key"].append(labels.get(kk, kk))
                    t["val"].append(ParamDesc.format_list(md_v, full=True))

            if in_jupyter:
                from IPython.display import HTML

                txt = ParamDesc._make_html_table(t, header=False)
                return HTML(txt)
            else:
                df = pd.DataFrame.from_dict(t)
                df = df.set_index("key")
                init_pandas_options()
                if not no_print:
                    print(df)
                return df

    @staticmethod
    def _make_html_table(d, header=True):
        if len(d) > 1:
            first_column_name = list(d.keys())[0]
            txt = """  
                <table>
                <tr>{}</tr>
                {}
                </table>""".format(
                "" if not header else "".join([f"<th>{k}</th>" for k in d.keys()]),
                "".join(
                    [
                        "<tr><th style='text-align: right;'>"
                        + d[first_column_name][i]
                        + "</th>"
                        + "".join(
                            [
                                f"<td style='text-align: left;'>{ParamDesc.format_list(d[k][i], full=True)}</td>"
                                for k in list(d.keys())[1:]
                            ]
                        )
                        + "</tr>"
                        for i in range(len(d[first_column_name]))
                    ]
                ),
            )
            return txt
        else:
            return ""

    @staticmethod
    def format_list(v, full=False):
        if isinstance(v, list):
            if full is True:
                return ",".join([str(x) for x in v])
            else:
                if len(v) == 1:
                    return v[0]
                if len(v) > 2:
                    return ",".join([str(x) for x in [v[0], v[1], "..."]])
                else:
                    return ",".join([str(x) for x in v])
        else:
            return v


class ParamNameDesc(ParamDesc):
    def __init__(self, name):
        super().__init__(name)
        self._short_name = name

    def load(self, db):
        md = {
            "typeOfLevel": [],
            "level": [],
            "date": [],
            "time": [],
            "step": [],
            "number": [],
            "paramId": [],
            "marsClass": [],
            "marsStream": [],
            "marsType": [],
            "experimentVersionNumber": [],
        }

        self.db = db
        self.md = {}
        self.levels = {}

        # print(f"par={par}")
        for b_name, b_df in db.blocks.items():
            if b_name == "scalar":
                q = f"shortName == '{self.short_name}'"
                dft = b_df.query(q)
            elif b_name == self.short_name:
                dft = b_df
            else:
                dft = None

            if dft is not None:
                for k in md.keys():
                    # print(f"{self.name}/{k}")
                    md[k].extend(dft[k].tolist())
                    # print(f"   df[{k}]={df[k]}")
            # print(df)

        self._parse(md)


class ParamIdDesc(ParamDesc):
    def __init__(self, param_id):
        super().__init__("")
        self._param_id = param_id

    def load(self, db):
        md = {
            "shortName": [],
            "typeOfLevel": [],
            "level": [],
            "date": [],
            "time": [],
            "step": [],
            "number": [],
            "paramId": [],
            "marsClass": [],
            "marsStream": [],
            "marsType": [],
            "experimentVersionNumber": [],
        }

        self.db = db
        self.md = {}
        self.levels = {}

        # print(f"par={par}"
        b_df = db.blocks.get("scalar", None)
        if b_df is not None:
            q = f"paramId == {self._param_id}"
            dft = b_df.query(q, engine="python")
            if dft is not None:
                for k in md.keys():
                    md[k].extend(dft[k].tolist())
                self._parse(md)



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import metview as mv
from metview import bindings
from metview.style import Style, Visdef


def test_visdef_object():
    vd = Visdef(
        "msymb",
        {
            "symbol_type": "text",
            "symbol_table_mode": "advanced",
            "symbol_advanced_table_text_font_colour": "black",
        },
    )

    # clone
    v = vd.clone()
    assert v.verb == vd.verb
    assert v.params == vd.params
    assert id(v) != id(vd)
    assert id(v.params) != id(vd.params)

    # change
    vd.change("msymb", "symbol_advanced_table_text_font_colour", "red")
    assert vd.params["symbol_advanced_table_text_font_colour"] == "red"
    vd.change("mcont", "symbol_advanced_table_text_font_colour", "blue")
    assert vd.params["symbol_advanced_table_text_font_colour"] == "red"

    # from request
    r = mv.msymb(
        {
            "symbol_type": "text",
            "symbol_table_mode": "advanced",
            "symbol_advanced_table_text_font_colour": "black",
        }
    )

    v = Visdef.from_request(r)
    assert v.verb == vd.verb
    for k, v in v.params.items():
        assert r[k.upper()].lower() == v.lower()


def test_style_object():
    vd = Visdef(
        "msymb",
        {
            "symbol_type": "text",
            "symbol_table_mode": "advanced",
            "symbol_advanced_table_text_font_colour": "black",
        },
    )

    # create
    s = Style("super", vd)
    assert s.name == "super"
    assert len(s.visdefs) == 1
    assert s.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "black"

    # update
    ns = s.update({"symbol_advanced_table_text_font_colour": "red"})
    assert id(ns) != id(s)
    assert s.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "black"
    assert ns.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "red"

    ns = s.update({"symbol_advanced_table_text_font_colour": "red"}, verb="mcont")
    assert id(ns) != id(s)
    assert s.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "black"
    assert ns.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "black"

    ns = s.update({"symbol_advanced_table_text_font_colour": "red"}, verb="msymb")
    assert id(ns) != id(s)
    assert s.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "black"
    assert ns.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "red"

    # multiple visdefs

    vd1 = Visdef(
        "msymb",
        {
            "symbol_type": "text",
            "symbol_table_mode": "advanced",
            "symbol_advanced_table_text_font_colour": "black",
        },
    )
    vd2 = Visdef("mgraph", {"graph_line_colour": "blue", "graph_line_thickness": 4})

    s = Style("super", [vd1, vd2])
    assert s.name == "super"
    assert len(s.visdefs) == 2
    assert s.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "black"
    assert s.visdefs[1].params["graph_line_colour"] == "blue"

    # update
    ns = s.update({"symbol_advanced_table_text_font_colour": "red"}, verb="msymb")
    assert id(ns) != id(s)
    assert s.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "black"
    assert ns.visdefs[0].params["symbol_advanced_table_text_font_colour"] == "red"

    ns = s.update(
        {
            "graph_line_colour": "yellow",
        },
        verb="mgraph",
    )
    assert id(ns) != id(s)
    assert s.visdefs[1].params["graph_line_colour"] == "blue"
    assert ns.visdefs[1].params["graph_line_colour"] == "yellow"


def test_style_set_grib_id():
    # mcont
    vd = Visdef(
        "mcont",
        {
            "legend": "on",
        },
    )

    s = Style("super", vd)
    ns = s.set_data_id("11")
    assert id(s) != id(ns)
    assert ns.visdefs[0].verb == "mcont"
    assert ns.visdefs[0].params["grib_id"] == "11"
    assert s.visdefs[0].params.get("grib_id", None) is None

    # mwind
    vd = Visdef(
        "mwind",
        {
            "legend": "on",
        },
    )

    s = Style("super", vd)
    ns = s.set_data_id("12")
    assert id(s) != id(ns)
    assert ns.visdefs[0].verb == "mwind"
    assert ns.visdefs[0].params["grib_id"] == "12"
    assert s.visdefs[0].params.get("grib_id", None) is None

    # other
    vd = Visdef(
        "msymb",
        {
            "legend": "on",
        },
    )

    s = Style("super", vd)
    ns = s.set_data_id("10")
    assert id(s) == id(ns)
    assert s.visdefs[0].params.get("grib_id", None) is None



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import math

import numpy as np
import pandas as pd

import metview as mv
from metview.layout import Layout
from metview.style import Visdef, Style, GeoView
from metview.title import Title
from metview.track import Track
from metview.scaling import Scaling

# logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")
# logging.basicConfig(level=logging.DEBUG, format="%(levelname)s - %(message)s")
# logging.basicConfig(level=logging.DEBUG)
LOG = logging.getLogger(__name__)


def _make_layers(*args, form_layout=False):
    # in the positional arguments we have two options:
    # 1. we only have non-list items. They belong to a single plot page.
    # 2. we only have list items. Each list item defines a separate plot page.
    plot_def = []
    if form_layout:
        plot_def = list(args)
        lst_cnt = sum([1 for x in plot_def if isinstance(x, list)])
        if not lst_cnt in [0, len(plot_def)]:
            raise Exception(
                f"Invalid plot arguments! Cannot mix list and non-list positional arguments."
            )

        if lst_cnt == 0:
            plot_def = [plot_def]
    else:
        plot_def = [list(args)]

    # plot def at this point a list of lists. Each list item describes a
    # map page item!
    layers = []
    for d in plot_def:
        data = None
        layer = []
        for item in d:
            if isinstance(item, (mv.Fieldset, Track)):
                data = item
                layer.append({"data": data, "vd": []})
            elif data is not None:
                assert len(layer) > 0
                layer[-1]["vd"].append(item)
        layers.append(layer)

    if form_layout:
        return layers
    else:
        return layers[0] if layers else []


def _make_visdef(
    data,
    vd,
    use_eccharts=False,
    style_db="param",
    plot_type="map",
    data_id=None,
    pos_values=None,
):
    if isinstance(data, mv.Fieldset):
        if len(vd) == 0:
            if use_eccharts:
                return [mv.style.make_eccharts_mcont()]
            else:
                vd = mv.style.get_db(name=style_db).visdef(
                    data, plot_type=plot_type, data_id=data_id
                )
                if plot_type == "diff" and pos_values is not None and pos_values:
                    s = mv.style.get_db(name=style_db).style(
                        data, plot_type=plot_type, data_id=data_id
                    )
                    if s is not None and len(s.visdefs) == 2:
                        neg_values = [-x for x in reversed(pos_values)]
                        s.visdefs[0].set_values_list(neg_values)
                        s.visdefs[1].set_values_list(pos_values)
                        vd = s.to_request()
                else:
                    vd = mv.style.get_db(name=style_db).visdef(
                        data, plot_type=plot_type, data_id=data_id
                    )
        else:
            for i, v in enumerate(vd):
                if isinstance(v, Style):
                    v = v.set_data_id(data_id)
                    vd[i] = v.to_request()
                elif (
                    isinstance(v, mv.Request) and data_id is not None and data_id != ""
                ):
                    v = Visdef.from_request(v)
                    v.set_data_id(data_id)
                    vd[i] = v.to_request()
        if vd is not None:
            return [x for x in vd if x is not None]
        else:
            return []
    else:
        return []


def _get_data_area(data):
    bb = []
    for d in data:
        b = d.bounding_box()
        if len(b) == 4:
            if math.fabs(b[1] - b[0]) > 160:
                b[0] = -90
                b[2] = 90
            if math.fabs(b[3] - b[1]) > 340:
                b[1] = -180
                b[3] = 180
        if len(bb) == 0:
            bb = b.tolist()
        else:
            bb = [
                min(b[0], bb[0]),
                min(b[1], bb[1]),
                max(b[2], bb[2]),
                max(b[3], bb[3]),
            ]
    return bb


def _make_view(view, area, plot_type="map", data=None):
    data = [] if data is None else data

    if area == "data" and data:
        area = _get_data_area(data)
        if len(area) != 4:
            area = "base"

    if view is not None:
        if area is not None:
            view = mv.make_geoview(area=area, style=view["coastlines"])
    else:
        if area is not None:
            view = mv.make_geoview(area=area, plot_type=plot_type)
        else:
            view = mv.make_geoview(area="base", plot_type=plot_type)
    return view


def _prepare_grid(d1, d2):
    if d1._db is not None and d2._db is not None:
        # interpolate from d2 to d1
        if d1._db.name in d2._db.regrid_from:
            # print(f"regrid: {d1.label} -> {d2.label}")
            # d11 = d1 + 0
            d = mv.regrid(data=d1, grid_definition_mode="template", template_data=d2[0])
            d = mv.grib_set_long(d, ["generatingProcessIdentifier", 148])
            d._db = d1._db._clone()
            # print(f"{ len(d)}")
            # print(f"{ len(d2)}")
            # print(" {}".format(mv.grib_get(d[0], ["numberOfDataPoints"])))
            # print(" {}".format(mv.grib_get(d2[0], ["numberOfDataPoints"])))
            return (d, d2)
        # interpolate from d1 to d2
        elif d2._db.name in d1._db.regrid_from:
            # print(f"regrid: {d2.label} -> {d1.label}")
            # d22 = d2 + 0
            d = mv.regrid(data=d2, grid_definition_mode="template", template_data=d1[0])
            d = mv.grib_set_long(d, ["generatingProcessIdentifier", 148])
            d._db = d2._db._clone()
            return (d1, d)

    return (d1, d2)


def _y_max(data):
    return max([max(d) for d in data])


def _y_min(data):
    return min([min(d) for d in data])


def _scale_data(data, style_db="param", plot_type="map"):
    if isinstance(data, mv.Fieldset):
        scaler = mv.style.get_db(name=style_db).units_scaler(data, plot_type=plot_type)
        if scaler is not None:
            return scaler.scale_value(data)
    return data


def _scale_xs(name, data):
    if isinstance(data, mv.Fieldset):
        scaler = None
        for f in data:
            if f.grib_get_string("shortName") != "lnsp":
                scaler = mv.style.get_db(name="param").units_scaler(f, plot_type="xs")
        if scaler is not None:
            r = mv.Fieldset()
            for f in data:
                if f.grib_get_string("shortName") != "lnsp":
                    r.append(scaler.scale_value(f))
                else:
                    r.append(f)
            return r
    return data


def plot_maps(
    *args,
    layout=None,
    view=None,
    area=None,
    use_eccharts=False,
    title_font_size=0.4,
    legend_font_size=0.35,
    frame=-1,
    animate="auto",
):
    """
    Plot maps with generic contents
    """

    # in the positional arguments we have two options:
    # 1. we only have non-list items. They belong to a single plot page.
    # 2. we only have list items. Each list item defines a separate plot page.
    plot_def = _make_layers(*args, form_layout=True)

    # collect the data items
    data_items = []
    for i, sc_def in enumerate(plot_def):
        for layer in sc_def:
            data = layer["data"]
            if isinstance(data, mv.Fieldset):
                data_items.append(data[0])

    # define the view
    view = _make_view(view, area, data=data_items)

    # build the layout
    num_plot = len(plot_def)
    dw = Layout().build_grid(page_num=num_plot, layout=layout, view=view)

    # the plot description
    desc = []

    title = Title(font_size=title_font_size)

    # build each scene
    data_id = ("d0", 0)
    for i, sc_def in enumerate(plot_def):
        desc.append(dw[i])
        # define layers
        data_items = []
        use_data_id = (
            sum([1 for layer in sc_def if isinstance(layer["data"], mv.Fieldset)]) > 1
        )
        for layer in sc_def:
            data = layer["data"]
            vd = layer["vd"]
            if isinstance(data, mv.Fieldset):
                if use_data_id:
                    data_items.append((data, data_id[0]))
                else:
                    data_items.append(data)
                if frame != -1:
                    if data.ds_param_info.scalar:
                        data = data[frame]
                    else:
                        data = data[2 * frame : 2 * frame + 2]

                data = _scale_data(data, style_db="param", plot_type="map")
            elif isinstance(data, Track):
                data = data.build(style=vd)

            desc.append(data)

            if isinstance(data, mv.Fieldset):
                vd = _make_visdef(
                    data,
                    vd,
                    use_eccharts=use_eccharts,
                    style_db="param",
                    plot_type="map",
                    data_id=data_id[0] if use_data_id else None,
                )
                if vd:
                    desc.extend(vd)

            data_id = (f"d{data_id[1]+1}", data_id[1] + 1)

        if data_items:
            legend = mv.mlegend(legend_text_font_size=legend_font_size)
            desc.append(legend)
            t = title.build(data_items)
            # LOG.debug(f"t={t}")
            desc.append(t)

    for i in range(len(plot_def), len(dw)):
        desc.append(dw[i])

    LOG.debug(f"desc={desc}")

    return mv.plot(desc, animate=animate)


def plot_diff_maps(
    *args,
    view=None,
    area=None,
    overlay=None,
    diff_style=None,
    pos_values=None,
    title_font_size=0.4,
    legend_font_size=0.35,
    frame=-1,
    animate="auto",
):
    """
    Plot difference maps
    """

    # handle default arguments
    pos_values = [] if pos_values is None else pos_values
    diff_style = [] if diff_style is None else diff_style
    if not isinstance(diff_style, list):
        diff_style = [diff_style]

    # define the view
    view = _make_view(view, area, plot_type="diff")

    # build the layout
    dw = Layout().build_diff(view=view)

    data = {}
    vd = {}

    # the positional arguments has the following order:
    # data1, visdef1.1 visdef1.2 ... data2, visdef2.1, visdef2.2 ...
    # the visdef objects are optional!
    assert len(args) >= 2
    assert isinstance(args[0], mv.Fieldset)
    layers = _make_layers(*args, form_layout=False)
    assert len(layers) == 2
    # LOG.debug(f"layers={layers}")
    data["0"] = layers[0]["data"]
    data["1"] = layers[1]["data"]
    vd["0"] = _make_visdef(data["0"], layers[0]["vd"])
    vd["1"] = _make_visdef(data["1"], layers[1]["vd"])

    # overlay data
    ov_data = {}
    ov_vd = {}
    if overlay is not None:
        # single value, list or tuple: a data item that will be plotted into each map
        if not isinstance(overlay, dict):
            if isinstance(overlay, tuple):
                ov_args = list(overlay)
            else:
                ov_args = [overlay] if not isinstance(overlay, list) else overlay
            # print(ov_args)
            ov_layers = _make_layers(*ov_args, form_layout=False)
            # print(ov_layers)
            assert len(ov_layers) == 1
            d = ov_layers[0]["data"]
            if isinstance(d, Track):
                d = d.build(style=ov_layers[0]["vd"])
            for k in ["d", "0", "1"]:
                ov_data[k] = d
                ov_vd[k] = _make_visdef(d, ov_layers[0]["vd"])
        else:
            pass

    # LOG.debug("len_0={}".format(len(data["0"])))
    # LOG.debug("len_1={}".format(len(data["0"])))

    # the plot description
    desc = []

    title = Title(font_size=title_font_size)

    # compute diff
    data["0"], data["1"] = _prepare_grid(data["0"], data["1"])
    data["d"] = data["0"] - data["1"]

    data["d"]._ds_param_info = data["1"].ds_param_info
    if data["0"].label and data["1"].label:
        data["d"]._label = "{}-{}".format(data["0"].label, data["1"].label)
    else:
        data["d"]._label = ""
    vd["d"] = _make_visdef(
        data["d"], diff_style, plot_type="diff", pos_values=pos_values
    )

    # LOG.debug("len_d={}".format(len(data["d"])))

    for i, k in enumerate(["d", "0", "1"]):
        desc.append(dw[i])
        if frame == -1:
            d = data[k]
        else:
            d = data[k][frame]
            d._ds_param_info = data[k]._ds_param_info
            d._label = data[k]._label

        desc.append(d)
        if vd[k]:
            desc.append(vd[k])

        # add overlay
        if k in ov_data:
            if isinstance(ov_data[k], mv.Fieldset):
                dd = ov_data[k] if frame == -1 else ov_data[k][frame]
            else:
                dd = ov_data[k]
            desc.append(dd)
            if k in ov_vd and ov_vd[k]:
                desc.append(ov_vd[k])

        t = title.build(data[k])
        legend = mv.mlegend(legend_text_font_size=legend_font_size)
        desc.append(legend)
        desc.append(t)

    # print(desc)
    return mv.plot(desc, animate=animate)


def plot_xs(
    *args,
    map_line=True,
    map_data=None,
    line=[],
    layout="",
    view=None,
    area=None,
    title_font_size=0.3,
    legend_font_size=0.2,
    frame=-1,
    animate="auto",
):
    """
    Plot cross section with map
    """

    assert len(line) == 4
    assert len(args) >= 1
    assert isinstance(args[0], mv.Fieldset)
    layers = _make_layers(*args, form_layout=False)
    assert len(layers) > 0

    # build the layout - if no map_data is specified no map view is
    # added to the layout
    if not map_line and map_data is None:
        view = None
    else:
        view = _make_view(view, area)
    dw = Layout().build_xs(line=line, map_view=view)

    # the plot description
    desc = []

    title = Title(font_size=title_font_size)
    data_items = []

    # build cross section plot
    desc.append(dw[0])
    for layer in layers:
        data = layer["data"]
        vd = _make_visdef(data, layer["vd"], plot_type="xs")
        param_info = data.ds_param_info
        # print(f"param_info={param_info}")
        data_items.append(data)
        # print(f"data={len(data)}")
        if param_info is not None and param_info.name == "wind3d":
            xs_d = mv.mcross_sect(
                data=data,
                line=line,
                wind_parallel="on",
                w_wind_scaling_factor_mode="compute",
                w_wind_scaling_factor="100",
                # bottom_level=1015,
                # top_level=250,
                # vertical_scaling = "log",
                # vertical_axis= vertical_axis
            )
            desc.append(xs_d)
        else:
            if param_info is not None:
                data = _scale_xs(param_info.name, data)
            desc.append(data)

        if vd:
            desc.extend(vd)
            # print(f"vd={vd}")

    t = title.build_xs(data_items)
    desc.append(t)

    # LOG.debug(f"desc={desc}")

    # build side map plot
    if map_line or map_data is not None:
        desc.append(dw[1])
        t = None
        if map_data is not None and len(map_data) > 0:
            layers = _make_layers(map_data, form_layout=False)
            data_items = []
            for layer in layers:
                data = layer["data"]
                vd = _make_visdef(data, layer["vd"])

                if isinstance(data, mv.Fieldset):
                    data_items.append(data)
                    if frame != -1:
                        data = data[frame]

                desc.append(data)
                if vd:
                    desc.extend(vd)

            if data_items:
                t = title.build(data_items)

        if map_line:
            # define xsection line graph
            graph = mv.mgraph(
                graph_line_colour="red", graph_line_thickness=3, graph_symbol="off"
            )

            lv = mv.input_visualiser(
                input_plot_type="geo_points",
                input_longitude_values=[line[1], line[3]],
                input_latitude_values=[line[0], line[2]],
            )

            desc.extend([lv, graph])

        if t is not None:
            desc.append(t)

    LOG.debug(f"desc={desc}")
    return mv.plot(desc, animate=animate)


def plot_stamp(
    *args,
    an=[],
    fc=[],
    layout=None,
    view=None,
    area=None,
    title_font_size=0.4,
    frame=-1,
    animate="auto",
    diff_base=None,
):
    """
    Plot ENS stamp maps
    """

    # define the view
    view = _make_view(view, area)

    desc = []
    data = {}
    vd = {}

    if diff_base is not None:
        assert isinstance(diff_base, mv.Fieldset)

    if len(args) > 0:
        assert isinstance(args[0], mv.Fieldset)
        layers = _make_layers(*args, form_layout=False)
        assert len(layers) == 1

        # prepare ens
        data["ens"] = layers[0]["data"]
        assert data["ens"] is not None
        if diff_base is not None:
            vd["ens"] = _make_visdef(data["ens"], [], style_db="diff")
        else:
            vd["ens"] = _make_visdef(data["ens"], layers[0]["vd"])

    # prepare an and fc
    d = {"an": an, "fc": fc}
    for k, v in d.items():
        if v:
            layers = _make_layers(v, form_layout=False)
            if layers:
                data[k] = layers[0]["data"]
                vd[k] = layers[0]["vd"]
                if diff_base is not None:
                    vd[k] = vd["ens"]
                else:
                    if len(vd[k]) == 0 and "ens" in vd:
                        vd[k] = vd["ens"]
                    else:
                        vd[k] = _make_visdef(data[k], vd[k])

    # determine ens number
    members = []
    if "ens" in data:
        members = data["ens"]._unique_metadata("number")
        LOG.debug(f"members={members}")
        if len(members) == 0:
            raise Exceception("No ENS data found in input!")

    # determine number of maps
    num = len(members) + sum([1 for x in ["an", "fc"] if x in data])

    # build the layout
    dw = Layout().build_stamp(num, layout=layout, view=view)

    if len(dw) < num + 1:
        raise Exception(f"Layout has less maps (={len(dw)}) than expected (={num})")

    title = Title(font_size=title_font_size)

    # ens members
    for i, m in enumerate(members):
        desc.append(dw[i])
        d = data["ens"].select(number=m)
        if diff_base is not None:
            d = d - diff_base
        desc.append(d)

        if vd["ens"]:
            desc.extend(vd["ens"])

        t = title.build_stamp(d, member=str(i))
        desc.append(t)

    # add an and fc
    n = len(members)
    for t in ["an", "fc"]:
        if t in data:
            desc.append(dw[n])
            d = data[t]
            if diff_base is not None:
                d = d - diff_base
            desc.append(d)
            if vd[t]:
                desc.append(vd[t])
            t = title.build_stamp(data[t], member="")
            desc.append(t)
            n += 1

    for i in range(n, len(dw)):
        desc.append(dw[i])

    if len(members) > 0 and "ens" in data:
        cont = mv.mcont(contour="off", contour_label="off")
        dummy = d = data["ens"].select(number=members[0])
        t = title.build(dummy)
        desc.extend([dw[-1], t, dummy, cont])

    return mv.plot(desc, animate=animate)


def plot_rmse(*args, ref=None, area=None, title_font_size=0.4, y_max=None):
    """
    Plot RMSE curve
    """

    desc = []

    if not isinstance(ref, mv.Fieldset):
        raise Exception(f"Missing or invalid ref argument!")

    layers = _make_layers(*args, form_layout=False)

    # compute the rmse for each input layer
    data = []  # list of tuples
    rmse_data = []
    title_data = []
    has_ef = False

    for layer in layers:
        if isinstance(layer["data"], mv.Fieldset):
            # determine ens number
            members = layer["data"]._unique_metadata("number")
            # print(f"members={members}")
            # ens forecast
            if len(members) > 1:
                if has_ef:
                    raise Exception("Only one ENS fieldset can be used in plot_rmse()!")
                has_ef = True
                em_d = None  # ens mean
                for m in members:
                    pf_d = layer["data"].select(number=m)
                    ref_d, pf_d = _prepare_grid(ref, pf_d)
                    data.append(("cf" if m == "0" else "pf", layer["data"]))
                    rmse_data.append(mv.sqrt(mv.average((pf_d - ref_d) ** 2)))
                    em_d = pf_d if em_d is None else em_d + pf_d

                # compute rmse for ens mean
                data.append(("em", layer["data"]))
                rmse_data.append(
                    mv.sqrt(mv.average((em_d / len(members) - ref_d) ** 2))
                )

            # deterministic forecast
            else:
                ref_d, dd = _prepare_grid(ref, layer["data"])
                data.append(("fc", layer["data"]))
                rmse_data.append(mv.sqrt(mv.average((dd - ref_d) ** 2)))

            title_data.append(layer["data"])

    # define x axis params
    dates = ref.valid_date()
    x_min = dates[0]
    x_max = dates[-1]
    x_tick = 1
    x_title = ""

    # define y axis params
    y_min = 0
    if y_max is None:
        y_tick, _, y_max = Layout.compute_axis_range(0, _y_max(rmse_data))
    else:
        y_tick, _, _ = Layout.compute_axis_range(0, y_max)
    y_title = "RMSE [" + mv.grib_get_string(ref[0], "units") + "]"

    # print(f"y_tick={y_tick} y_max={y_max}")

    # define the view
    view = Layout().build_rmse(
        x_min, x_max, y_min, y_max, x_tick, y_tick, x_title, y_title
    )
    desc.append(view)

    # define curves
    ef_label = {"cf": "ENS cf", "pf": "ENS pf", "em": "ENS mean"}
    ef_colour = {"cf": "black", "pf": "red", "em": "kelly_green"}
    fc_colour = ["red", "blue", "green", "black", "cyan", "evergreen", "gold", "pink"]
    if has_ef:
        fc_colour = [x for x in fc_colour if x not in list(ef_colour.values())]

    pf_label_added = False
    colour_idx = -1
    legend_item_count = 0
    for i, d in enumerate(rmse_data):
        vis = mv.input_visualiser(
            input_x_type="date", input_date_x_values=dates, input_y_values=d
        )

        vd = {"graph_type": "curve"}
        line_colour = "black"
        line_width = 1

        if data[i][0] == "fc":
            line_width = 3
            colour_idx = (colour_idx + 1) % len(fc_colour)
            line_colour = fc_colour[colour_idx]
            # print(f"label={data[i][1][0].label}")
            vd["legend_user_text"] = data[i][1].label
            vd["legend"] = "on"
            legend_item_count += 1
        elif data[i][0] == "pf":
            line_width = 1
            line_colour = ef_colour["pf"]
            if not pf_label_added:
                pf_label_added = True
                vd["legend_user_text"] = ef_label.get("pf", "")
                vd["legend"] = "on"
                legend_item_count += 1
        elif data[i][0] in ["cf", "em"]:
            line_width = 3
            line_colour = ef_colour[data[i][0]]
            vd["legend_user_text"] = ef_label.get(data[i][0], "")
            vd["legend"] = "on"
            legend_item_count += 1

        vd["graph_line_colour"] = line_colour
        vd["graph_line_thickness"] = line_width

        desc.append(vis)
        desc.append(mv.mgraph(**vd))

    # add title
    title = Title(font_size=title_font_size)
    t = title.build_rmse(ref, title_data)
    if t is not None:
        desc.append(t)

    # add legend
    leg_left = 3.5
    # legY = 14
    leg_height = legend_item_count * (0.35 + 0.5) + (legend_item_count + 1) * 0.1
    leg_bottom = 17.5 - leg_height

    # Legend
    legend = mv.mlegend(
        legend_display_type="disjoint",
        legend_entry_plot_direction="column",  # "row",
        legend_text_composition="user_text_only",
        legend_border="on",
        legend_border_colour="black",
        legend_box_mode="positional",
        legend_box_x_position=leg_left,
        legend_box_y_position=leg_bottom,
        legend_box_x_length=4,
        legend_box_y_length=leg_height,
        legend_text_font_size=0.35,
        legend_box_blanking="on",
    )
    desc.append(legend)

    mv.plot(desc, animate=False)


def plot_cdf(*args, location=None, title_font_size=0.4, x_range=None):
    """
    Plot CDF curve
    """

    # check x range
    x_range = [] if x_range is None else x_range
    if x_range and len(x_range) not in [2, 3]:
        raise Exception(
            f"plot_cdf: invalid x_range specified. Format [x_min, x_max, [x_tick]]"
        )
    if len(x_range) == 2 and x_range[1] <= x_range[0]:
        raise Exception(
            f"plot_cdf: invalid x_range specified. x_min={x_range[0]} >= x_max={x_range[1]}"
        )

    layers = _make_layers(*args, form_layout=False)

    desc = []
    cdf_data = []
    cdf_label = []
    title_data = []
    y_values = np.arange(0, 101)
    plot_units = ""
    units_scaler = None

    # compute the cdf for each input layer
    for layer in layers:
        if isinstance(layer["data"], mv.Fieldset):
            # we assume each field has the same units and paramId
            if plot_units == "":
                meta = mv.grib_get(layer["data"][0], ["units", "paramId"])
                if meta and len(meta[0]) == 2:
                    meta = {"units": meta[0][0], "paramId": meta[0][1]}
                    units_scaler = Scaling.find_item(meta)
                    if units_scaler is not None:
                        plot_units = units_scaler.to_units
                    else:
                        plot_units = meta.get("units", "")

            # determine ens number and steps
            members = layer["data"]._unique_metadata("number")
            steps = layer["data"]._unique_metadata("step")
            # print(f"members={members}")
            # ens forecast
            if len(members) > 1:
                for step in steps:
                    v = layer["data"].select(step=step)
                    v = mv.nearest_gridpoint(v, location)
                    # print(f"step={step}")
                    x = np.percentile(v, y_values)
                    if units_scaler is not None:
                        x = units_scaler.scale_value(x)
                    # print(f" x={x}")
                    cdf_data.append(x)
                    cdf_label.append(layer["data"].label + f" +{step}h")

            # deterministic forecast
            else:
                raise Exception(f"plot_cds: only ENS data accepted as input!")

            title_data.append(layer["data"])

    # define x axis params
    if not x_range:
        x_tick, x_min, x_max = Layout.compute_axis_range(
            _y_min(cdf_data), _y_max(cdf_data)
        )
    elif len(x_range) == 2:
        x_min = x_range[0]
        x_max = x_range[1]
        x_tick, _, _ = Layout.compute_axis_range(x_min, x_max)
    elif len(x_range) == 3:
        x_min = x_range[0]
        x_max = x_range[1]
        x_tick = x_range[2]
    else:
        raise Exception(f"plot_cdf: invalid x_range={x_range} specified!")

    # print(f"x_tick={x_tick} x_min={x_min} x_max={x_max}")
    x_title = f"[{plot_units}]"

    # define y axis params
    y_min = 0
    y_max = 100
    y_tick = 10
    y_title = "Percentage [%]"

    # define the view
    view = Layout().build_xy(
        x_min, x_max, y_min, y_max, x_tick, y_tick, x_title, y_title
    )
    desc.append(view)

    # define curves
    line_colours = [
        "red",
        "blue",
        "green",
        "black",
        "cyan",
        "evergreen",
        "gold",
        "pink",
    ]
    line_styles = ["solid", "dash", "dotted"]

    colour_idx = -1
    style_idx = 0

    for i, d in enumerate(cdf_data):
        vis = mv.input_visualiser(input_x_values=d, input_y_values=y_values)
        colour_idx = (colour_idx + 1) % len(line_colours)

        vd = mv.mgraph(
            graph_type="curve",
            graph_line_colour=line_colours[colour_idx],
            graph_line_thickness=3,
            legend_user_text=cdf_label[i],
            legend="on",
        )

        desc.append(vis)
        desc.append(vd)

    # add title
    title = Title(font_size=title_font_size)
    t = title.build_cdf(title_data)
    if t is not None:
        desc.append(t)

    # add legend
    legX = 3.5
    legY = 14

    # Legend
    legend = mv.mlegend(
        legend_display_type="disjoint",
        legend_entry_plot_direction="column",
        legend_text_composition="user_text_only",
        legend_border="on",
        legend_border_colour="black",
        legend_box_mode="positional",
        legend_box_x_position=legX,
        legend_box_y_position=legY,
        legend_box_x_length=4,
        legend_box_y_length=3,
        legend_text_font_size=0.35,
        legend_box_blanking="on",
    )
    desc.append(legend)

    mv.plot(desc, animate=False)


def plot_xs_avg(
    *args,
    area=None,
    direction="ew",
    vertical_scaling="linear",
    title_font_size=0.3,
    legend_font_size=0.2,
    axis_font_size=0.3,
    animate="auto",
):
    """
    Plot average cross section
    """

    if area is None:
        area = [90, -180, -90, 180]

    assert isinstance(args[0], mv.Fieldset)
    layers = _make_layers(*args, form_layout=False)
    assert len(layers) > 0

    view = Layout().build_xs_avg(
        area=area,
        direction=direction,
        bottom_level=1000,
        top_level=50,
        vertical_scaling=vertical_scaling,
        axis_label_height=axis_font_size,
    )

    # the plot description
    desc = []

    title = Title(font_size=title_font_size)
    data_items = []

    # build cross section plot
    desc.append(view)
    for layer in layers:
        data = layer["data"]

        vd = _make_visdef(data, layer["vd"], plot_type="xs")
        param_info = data.ds_param_info
        # print(f"param_info={param_info}")
        data_items.append(data)
        # print(f"data={len(data)}")

        if param_info is not None:
            data = _scale_xs(param_info.name, data)
        desc.append(data)

        if vd:
            desc.extend(vd)
        # print(f"vd={vd}")

    t = title.build_xs(data_items)
    desc.append(t)

    legend = mv.mlegend(legend_text_font_size=legend_font_size)
    desc.append(legend)
    desc.append(t)

    LOG.debug(f"desc={desc}")

    return mv.plot(desc, animate=animate)



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import os
import sys

import yaml

LOG = logging.getLogger(__name__)

ETC_PATH = os.path.join(os.path.dirname(__file__), "etc")


class UnitsScalingMethod:
    """
    Performs units scaling of values
    """

    def __init__(self, scaling=1.0, offset=0.0, from_units="", to_units=""):
        self.scaling = scaling
        self.offset = offset
        self.from_units = from_units
        self.to_units = to_units

    def scale_value(self, value):
        return self.scaling * value + self.offset

    def inverse_scale_value(self, value):
        return (value - self.offset) / self.scaling

    def need_scaling(self, meta, scaling_retrieved, scaling_derived):
        gen_id = meta.get("generatingProcessIdentifier", 0)
        try:
            gen_id = int(gen_id)
        except:
            gen_id = 0

        return (scaling_retrieved and gen_id != 254) or (
            scaling_derived and gen_id == 254
        )

    def __str__(self):
        return "scaling: {} offset:{} from_units:{} to_units:{}".format(
            self.scaling, self.offset, self.from_units, self.to_units
        )


class ScalingRule:
    """
    Defines what countour scaling should be applied to a given field based
    on its metadata
    """

    def __init__(self, to_units, conf):
        self.to_units = to_units
        self.units_methods = [
            it for it in Scaling.methods if it.to_units == self.to_units
        ]
        self.conf = conf

    def find_method(self, meta):
        from_units = meta.get("units", "")
        if from_units == "":
            return None

        method = None
        for item in self.units_methods:
            if item.from_units == from_units:
                method = item
                break

        if method is None:
            return None

        for m in self.conf.get("match", []):
            match = False
            for k, v in m.items():
                if meta.get(k, None) == v:
                    match = True
                else:
                    break

            if match:
                return method

        param_id = meta.get("paramId", "")
        if param_id and param_id in self.conf.get("paramId", []):
            return method

        short_name = meta.get("shortName", "")
        if short_name and short_name in self.conf.get("shortName", []):
            return method

        return None

    def __str__(self):
        return "to_units:{}".format(self.to_units)


class Scaling:
    methods = []
    rules = []
    loaded = False

    @staticmethod
    def find_item(meta):
        if not Scaling.loaded:
            Scaling._load_def()
            Scaling.loaded = True

        for item in Scaling.rules:
            d = item.find_method(meta)
            if d:
                return d
        return None

    @staticmethod
    def _load_def():
        # load units conversion definition
        file_name = os.path.join(ETC_PATH, "units-rules.json")
        with open(file_name) as f:
            data = json.load(f)
            for k, v in data.items():
                for item in v:
                    item["to_units"] = item.pop("to")
                    item["from_units"] = item.pop("from")
                    Scaling.methods.append(UnitsScalingMethod(**item))

        # load rules defining when to apply scaling on a parameter
        file_name = os.path.join(ETC_PATH, "scaling_ecmwf.yaml")
        # print(f"file_name={file_name}")
        with open(file_name) as f:
            data = yaml.load(f, Loader=yaml.SafeLoader)
            for to_units, item in data.items():
                Scaling.rules.append(ScalingRule(to_units, item))



#
# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

"""
ipython is not None when running a notebook
"""

import logging
import sys


ipython_active = None


def is_ipython_active():
    global ipython_active
    if ipython_active is None:
        try:
            from IPython import get_ipython

            ipython_active = get_ipython() is not None
        except Exception:
            ipython_active = False
    return ipython_active


def import_widgets():
    try:
        widgets = __import__("ipywidgets", globals(), locals())
        return widgets
    except ImportError as imperr:
        print("Could not import ipywidgets module - animation widget will not appear")
        print("Call setoutput('jupyter', plot_widget=False) to suppress this message")
        return None



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
import zipfile

import metview as mv


def load_dataset(filename, check_local=False):
    def _simple_download(url, target):
        import requests

        r = requests.get(url, allow_redirects=True)
        r.raise_for_status()
        open(target, "wb").write(r.content)

    if check_local and os.path.exists(filename):
        try:
            return mv.read(filename)
        except:
            return None

    base_url = "https://get.ecmwf.int/test-data/metview/gallery/"
    try:
        # d = mv.download(url=base_url + filename, target=filename)
        _simple_download(os.path.join(base_url, filename), filename)
    except Exception as e:
        raise Exception(
            f"Could not download file={filename} from the download server. {e}"
        )

    d = None
    if filename.endswith(".zip"):
        with zipfile.ZipFile(filename, "r") as f:
            f.extractall()
    else:
        try:
            d = mv.read(filename)
        except:
            pass
    return d



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from inspect import signature
import sys

import numpy as np
import eccodes

from . import maths
from .temporary import temp_file

from . import indexdb as indexdb
from . import utils


BITS_PER_VALUE_FOR_WRITING = 24


class CodesHandle:
    """Wraps an ecCodes handle"""

    MISSING_VALUE = 1e34

    STRING = ("s", eccodes.codes_get_string, eccodes.codes_set_string)
    LONG = ("l", eccodes.codes_get_long, eccodes.codes_set_long)
    DOUBLE = ("d", eccodes.codes_get_double, eccodes.codes_set_double)
    LONG_ARRAY = ("la", eccodes.codes_get_long_array, eccodes.codes_set_long_array)
    DOUBLE_ARRAY = (
        "da",
        eccodes.codes_get_double_array,
        eccodes.codes_set_double_array,
    )
    NATIVE = ("n", eccodes.codes_get, eccodes.codes_set)
    NATIVE_ARRAY = ("na", eccodes.codes_get_array, eccodes.codes_set_array)

    _STR_TYPES = {
        v[0]: tuple(v[1:3])
        for v in [STRING, LONG, DOUBLE, LONG_ARRAY, DOUBLE_ARRAY, NATIVE, NATIVE_ARRAY]
    }

    def __init__(self, handle, path, offset):
        self.handle = handle
        self.path = path
        self.offset = offset
        eccodes.codes_set(handle, "missingValue", CodesHandle.MISSING_VALUE)

    def __del__(self):
        # print("CodesHandle:release ", self)
        eccodes.codes_release(self.handle)

    def clone(self):
        new_handle = eccodes.codes_clone(self.handle)
        return CodesHandle(new_handle, None, None)

    # def __str__(self):
    #    s = "CodesHandle("
    #    return s + str(self.handle) + "," + self.path + "," + str(self.offset) + ")"

    def call_get_func(self, func, key, default=False):
        try:
            result = func(self.handle, key)
        except Exception as exp:
            if default != False:
                return default
            else:
                raise exp
        return result

    def get_any(self, keys, key_type=None):
        # single key with key_type specified
        if key_type is not None:
            assert isinstance(keys, str)
            func = key_type[1]
            return func(self.handle, keys)
        # list of keys
        else:
            result = []
            for key in keys:
                key_type_str = "s"
                parts = key.split(":")
                if len(parts) == 2:
                    key, key_type_str = parts
                    if (
                        key_type_str == "n"
                        and eccodes.codes_get_size(self.handle, key) > 1
                    ):
                        key_type_str = "na"
                func = CodesHandle._STR_TYPES.get(key_type_str, None)[0]
                result.append(self.call_get_func(func, key, default=None))
            return result

    def get_string(self, key):
        return eccodes.codes_get_string(self.handle, key)

    def get_long(self, key):
        return eccodes.codes_get_long(self.handle, key)

    def get_double(self, key):
        return eccodes.codes_get_double(self.handle, key)

    def get_long_array(self, key):
        return eccodes.codes_get_long_array(self.handle, key)

    def get_double_array(self, key):
        return eccodes.codes_get_double_array(self.handle, key)

    def get_values(self):
        vals = eccodes.codes_get_values(self.handle)
        if self.get_long("bitmapPresent"):
            vals[vals == CodesHandle.MISSING_VALUE] = np.nan
        return vals

    def set_any(self, keys_and_vals, key_type=None):
        for key, v in zip(keys_and_vals[0::2], keys_and_vals[1::2]):
            if key_type is not None:
                func = key_type[2]
            else:
                key_type_str = "s"  # default is str
                parts = key.split(":")
                if len(parts) == 2:
                    key, key_type_str = parts
                func = CodesHandle._STR_TYPES.get(key_type_str, None)[1]
            func(self.handle, key, v)

    def set_string(self, key, value):
        eccodes.codes_set_string(self.handle, key, value)

    def set_long(self, key, value):
        eccodes.codes_set_long(self.handle, key, value)

    def set_double(self, key, value):
        eccodes.codes_set_double(self.handle, key, value)

    def set_double_array(self, key, value):
        eccodes.codes_set_double_array(self.handle, key, value)

    def set_values(self, value):
        # replace nans with missing values
        values_nans_replaced = np.nan_to_num(
            value, copy=True, nan=CodesHandle.MISSING_VALUE
        )
        self.set_long("bitsPerValue", BITS_PER_VALUE_FOR_WRITING)
        self.set_double_array("values", values_nans_replaced)

    def write(self, fout, path):
        self.offset = fout.tell()
        eccodes.codes_write(self.handle, fout)
        if path:
            self.path = path


class GribFile:
    """Encapsulates a GRIB file, giving access to an iteration of CodesHandles"""

    def __init__(self, path):
        self.path = path
        self.file = open(path, "rb")
        self.num_messages = eccodes.codes_count_in_file(self.file)

    def __del__(self):
        try:
            # print("GribFile:close")
            self.file.close()
        except Exception:
            pass

    def __len__(self):
        return self.num_messages

    def __iter__(self):
        return self

    def __next__(self):
        handle = self._next_handle()
        if handle is None:
            raise StopIteration()
        return handle

    def _next_handle(self):
        offset = self.file.tell()
        handle = eccodes.codes_new_from_file(self.file, eccodes.CODES_PRODUCT_GRIB)
        if not handle:
            return None
        return CodesHandle(handle, self.path, offset)


class Field:
    """Encapsulates single GRIB message"""

    def __init__(self, handle, gribfile, keep_values_in_memory=False, temp=None):
        self.handle = handle
        self.gribfile = gribfile
        self.temp = temp
        self.vals = None
        self.keep_values_in_memory = keep_values_in_memory

    def grib_get(self, *args, **kwargs):
        return self.handle.get_any(*args, **kwargs)

    def values(self):
        if self.vals is None:
            vals = self.handle.get_values()
            if self.keep_values_in_memory:
                self.vals = vals
        else:
            vals = self.vals
        return vals

    def latitudes(self):
        return self.handle.get_double_array("latitudes")

    def longitudes(self):
        return self.handle.get_double_array("longitudes")

    def grib_set(self, *args, **kwargs):
        result = self.clone()
        result.handle.set_any(*args, **kwargs)
        return result

    def encode_values(self, value):
        self.handle.set_long("bitmapPresent", 1)
        self.handle.set_values(value)
        if not self.keep_values_in_memory:
            self.vals = None

    def write(self, fout, path, temp=None):
        self.temp = temp  # store a reference to the temp file object for persistence
        self.handle.write(fout, path)

    def grib_index(self):
        return (self.handle.path, self.handle.offset)
        # return (self.handle.path, self.grib_get("offset", key_type=CodesHandle.LONG))

    def clone(self):
        c = Field(
            self.handle.clone(),
            self.gribfile,
            self.keep_values_in_memory,
        )
        c.vals = None
        return c

    def field_func(self, func):
        """Applies a function to all values, returning a new Field"""
        result = self.clone()
        result.vals = func(self.values())
        result.encode_values(result.vals)
        return result

    def field_other_func(self, func, other, reverse_args=False):
        """Applies a function with something to all values, returning a new Field"""
        result = self.clone()
        if isinstance(other, Field):
            other = other.values()
        if reverse_args:
            result.vals = func(other, self.values())
        else:
            result.vals = func(self.values(), other)
        result.encode_values(result.vals)
        return result


# decorator to implement math functions in Fieldset
def wrap_maths(cls):
    def wrap_single_method(fn):
        def wrapper(self):
            return self.field_func(fn)

        return wrapper

    def wrap_double_method(fn, **opt):
        def wrapper(self, *args):
            return self.fieldset_other_func(fn, *args, **opt)

        return wrapper

    for name, it in cls.WRAP_MATHS_ATTRS.items():
        if not isinstance(it, tuple):
            it = (it, {})
        fn, opt = it
        n = len(signature(fn).parameters)
        if n == 1:
            setattr(cls, name, wrap_single_method(fn))
        elif n == 2:
            setattr(cls, name, wrap_double_method(fn, **opt))
    return cls


@wrap_maths
class Fieldset:
    """A set of Fields, each of which can come from different GRIB files"""

    WRAP_MATHS_ATTRS = {
        "abs": maths.abs,
        "acos": maths.acos,
        "asin": maths.asin,
        "atan": maths.atan,
        "atan2": maths.atan2,
        "cos": maths.cos,
        "div": maths.floor_div,
        "exp": maths.exp,
        "log": maths.log,
        "log10": maths.log10,
        "mod": maths.mod,
        "sgn": maths.sgn,
        "sin": maths.sin,
        "square": maths.square,
        "sqrt": maths.sqrt,
        "tan": maths.tan,
        "__neg__": maths.neg,
        "__pos__": maths.pos,
        "__invert__": maths.not_func,
        "__add__": maths.add,
        "__radd__": (maths.add, {"reverse_args": True}),
        "__sub__": maths.sub,
        "__rsub__": (maths.sub, {"reverse_args": True}),
        "__mul__": maths.mul,
        "__rmul__": (maths.mul, {"reverse_args": True}),
        "__truediv__": maths.div,
        "__rtruediv__": (maths.div, {"reverse_args": True}),
        "__pow__": maths.pow,
        "__rpow__": (maths.pow, {"reverse_args": True}),
        "__ge__": maths.ge,
        "__gt__": maths.gt,
        "__le__": maths.le,
        "__lt__": maths.lt,
        "__eq__": maths.eq,
        "__ne__": maths.ne,
        "__and__": maths.and_func,
        "__or__": maths.or_func,
        "bitmap": (maths.bitmap, {"use_first_from_other": True}),
        "nobitmap": maths.nobitmap,
    }

    # QUALIFIER_MAP = {"float": "d"}

    # INT_KEYS = ["Nx", "Ny", "number"]

    def __init__(
        self, path=None, fields=None, keep_values_in_memory=False, temporary=False
    ):
        self.fields = []
        self.count = 0
        self.temporary = None
        self._db = None

        if (path is not None) and (fields is not None):
            raise ValueError("Fieldset cannot take both path and fields")

        if path:
            if isinstance(path, list):
                v = []
                for p in path:
                    v.extend(utils.get_file_list(p))
                path = v
            else:
                path = utils.get_file_list(path)

            for p in path:
                g = GribFile(p)
                self.count = len(g)
                for handle in g:
                    self.fields.append(Field(handle, p, keep_values_in_memory))
        if temporary:
            self.temporary = temp_file()
        if fields:
            self.fields = fields

    def __len__(self):
        return len(self.fields)

    def __str__(self):
        n = len(self)
        s = "s" if n > 1 else ""
        return f"Fieldset ({n} field{s})"

    def _grib_get(self, *args, as_list=False, **kwargs):
        ret = [x.grib_get(*args, **kwargs) for x in self.fields]
        return ret if as_list else Fieldset._list_or_single(ret)

    def grib_get_string(self, key):
        return self._grib_get(key, key_type=CodesHandle.STRING)

    def grib_get_long(self, key):
        return self._grib_get(key, key_type=CodesHandle.LONG)

    def grib_get_double(self, key):
        return self._grib_get(key, key_type=CodesHandle.DOUBLE)

    def grib_get_long_array(self, key):
        return self._grib_get(key, key_type=CodesHandle.LONG_ARRAY)

    def grib_get_double_array(self, key):
        return self._grib_get(key, key_type=CodesHandle.DOUBLE_ARRAY)

    def grib_get(self, keys, grouping="field"):
        if grouping not in ["field", "key"]:
            raise ValueError(f"grib_get: grouping must be field or key, not {grouping}")
        ret = self._grib_get(keys, as_list=True)
        if grouping == "key":
            ret = list(map(list, zip(*ret)))  # transpose lists of lists
        return ret

    def _grib_set(self, *args, **kwargs):
        result = Fieldset(temporary=True)
        path = result.temporary.path
        with open(path, "wb") as fout:
            for f in self.fields:
                result._append_field(f.grib_set(*args, **kwargs))
                result.fields[-1].write(fout, path, temp=result.temporary)
        return result

    def grib_set_string(self, keys_and_vals):
        return self._grib_set(keys_and_vals, key_type=CodesHandle.STRING)

    def grib_set_long(self, keys_and_vals):
        return self._grib_set(keys_and_vals, key_type=CodesHandle.LONG)

    def grib_set_double(self, keys_and_vals):
        return self._grib_set(keys_and_vals, key_type=CodesHandle.DOUBLE)

    def grib_set(self, keys_and_vals):
        return self._grib_set(keys_and_vals)

    # def grib_set_double_array(self, key, value):
    #    return self._grib_set_any(key, value, "grib_set_double_array")

    def values(self):
        v = [x.values() for x in self.fields]
        return self._make_2d_array(v)

    def set_values(self, values):
        if isinstance(values, list):
            list_of_arrays = values
        else:
            if len(values.shape) > 1:
                list_of_arrays = [a for a in values]
            else:
                list_of_arrays = [values]
        if len(list_of_arrays) != len(self.fields):
            msg = str(len(list_of_arrays)) + " instead of " + str(len(self.fields))
            raise ValueError("set_values has the wrong number of arrays: " + msg)

        return self.fieldset_other_func(
            maths.set_from_other, list_of_arrays, index_other=True
        )

    def write(self, path):
        with open(path, "wb") as fout:
            for f in self.fields:
                f.write(fout, path)

    def grib_index(self):
        return [i.grib_index() for i in self.fields]

    def _append_field(self, field):
        self.fields.append(field)
        self._db = None

    def __getitem__(self, index):
        try:
            if isinstance(index, np.ndarray):
                return Fieldset(fields=[self.fields[i] for i in index])
            # # GRIB key
            # elif isinstance(index, str):
            #     keyname = index
            #     qualifier = "n"
            #     parts = keyname.split(":")
            #     if len(parts) > 1:
            #         keyname = parts[0]
            #         qualifier = Fieldset.QUALIFIER_MAP[parts[1]]

            #     native_key = keyname + ":" + qualifier
            #     # print("native_key", native_key)
            #     value = self.grib_get([native_key])[0][0]
            #     # print(value)
            #     if value is None:
            #         raise IndexError
            #     if index in Fieldset.INT_KEYS:
            #         # print("int value:", int(value))
            #         return int(value)
            #     # if isinstance(value, float):
            #     #    return int(value)
            #     return value
            elif isinstance(index, str):
                return self._get_db().select_with_name(index)
            else:
                return Fieldset(fields=self._always_list(self.fields[index]))
        except IndexError as ide:
            # print("This Fieldset contains", len(self), "fields; index is", index)
            raise ide

    def append(self, other):
        self.fields = self.fields + other.fields
        self._db = None

    def merge(self, other):
        result = Fieldset(fields=self.fields, temporary=True)
        result.append(other)
        return result

    # def items(self):
    #     its = []
    #     for i in range(len(self)):
    #         its.append((i, Fieldset(fields=[self.fields[i]])))
    #     return its

    # def to_dataset(self, via_file=True, **kwarg):
    #     # soft dependency on cfgrib
    #     try:
    #         import xarray as xr
    #     except ImportError:  # pragma: no cover
    #         print("Package xarray not found. Try running 'pip install xarray'.")
    #         raise
    #     # dataset = xr.open_dataset(self.url(), engine="cfgrib", backend_kwargs=kwarg)
    #     if via_file:
    #         dataset = xr.open_dataset(self.url(), engine="cfgrib", backend_kwargs=kwarg)
    #     else:
    #         print("Using experimental cfgrib interface to go directly from Fieldset")
    #         dataset = xr.open_dataset(self, engine="cfgrib", backend_kwargs=kwarg)
    #     return dataset

    def field_func(self, func):
        """Applies a function to all values in all fields"""
        result = Fieldset(temporary=True)
        with open(result.temporary.path, "wb") as fout:
            for f in self.fields:
                result._append_field(f.field_func(func))
                result.fields[-1].write(
                    fout, result.temporary.path, temp=result.temporary
                )
        return result

    def fieldset_other_func(
        self,
        func,
        other,
        reverse_args=False,
        index_other=False,
        use_first_from_other=False,
    ):
        """Applies a function to a fieldset and a scalar/fieldset, e.g. F+5"""
        # print(
        #     f"Fieldset.fieldset_other_func() func={func}, other={other}, reverse_args={reverse_args}, index_other={index_other}, use_first_from_other={use_first_from_other}"
        # )

        def _process_one(f, g, result):
            new_field = f.field_other_func(func, g, reverse_args=reverse_args)
            result._append_field(new_field)
            result.fields[-1].write(fout, result.temporary.path, temp=result.temporary)

        result = Fieldset(temporary=True)
        with open(result.temporary.path, "wb") as fout:
            if isinstance(other, Fieldset):
                if len(other) == len(self.fields):
                    for f, g in zip(self.fields, other.fields):
                        _process_one(f, g, result)
                elif use_first_from_other:
                    for f in self.fields:
                        _process_one(f, other.fields[0], result)
                else:
                    raise Exception(
                        f"Fieldsets must have the same number of fields for this operation! {len(self.fields)} != {len(other)}"
                    )
            elif index_other:
                for f, g in zip(self.fields, other):
                    _process_one(f, g, result)
            else:
                for f in self.fields:
                    _process_one(f, other, result)
        return result

    def base_date(self):
        if len(self.fields) > 0:
            result = []
            for f in self.fields:
                md = f.grib_get(["dataDate", "dataTime"])
                result.append(
                    utils.date_from_ecc_keys(md[0], md[1]) if len(md) == 2 else None
                )
            return Fieldset._list_or_single(result)

    def valid_date(self):
        if len(self.fields) > 0:
            result = []
            for f in self.fields:
                md = f.grib_get(["validityDate", "validityTime"])
                result.append(
                    utils.date_from_ecc_keys(md[0], md[1]) if len(md) == 2 else None
                )
            return Fieldset._list_or_single(result)

    def accumulate(self):
        if len(self.fields) > 0:
            result = [np.nan] * len(self.fields)
            for i, f in enumerate(self.fields):
                result[i] = np.sum(f.values())
            return Fieldset._list_or_single(result)

    def average(self):
        if len(self.fields) > 0:
            result = [np.nan] * len(self.fields)
            for i, f in enumerate(self.fields):
                result[i] = f.values().mean()
            return Fieldset._list_or_single(result)

    def maxvalue(self):
        if len(self.fields) > 0:
            result = np.array([np.nan] * len(self.fields))
            for i, f in enumerate(self.fields):
                result[i] = f.values().max()
            return result.max()

    def minvalue(self):
        if len(self.fields) > 0:
            result = np.array([np.nan] * len(self.fields))
            for i, f in enumerate(self.fields):
                result[i] = f.values().min()
            return result.min()

    def _make_single_result(self, v):
        assert len(self) > 0
        result = Fieldset(temporary=True)
        with open(result.temporary.path, "wb") as fout:
            f = self.fields[0].clone()
            f.encode_values(v)
            result._append_field(f)
            result.fields[-1].write(fout, result.temporary.path, temp=result.temporary)
        return result

    def mean(self):
        if len(self.fields) > 0:
            v = self.fields[0].values()
            for i in range(1, len(self.fields)):
                v += self.fields[i].values()
            v = v / len(self.fields)
            return self._make_single_result(v)
        else:
            return None

    def rms(self):
        if len(self.fields) > 0:
            v = np.square(self.fields[0].values())
            for i in range(1, len(self.fields)):
                v += np.square(self.fields[i].values())
            v = np.sqrt(v / len(self.fields))
            return self._make_single_result(v)
        else:
            return None

    def stdev(self):
        v = self._compute_var()
        return self._make_single_result(np.sqrt(v)) if v is not None else None

    def sum(self):
        if len(self.fields) > 0:
            v = self.fields[0].values()
            for i in range(1, len(self.fields)):
                v += self.fields[i].values()
            return self._make_single_result(v)
        else:
            return None

    def var(self):
        v = self._compute_var()
        return self._make_single_result(v) if v is not None else None

    def _compute_var(self):
        if len(self.fields) > 0:
            v2 = self.fields[0].values()
            v1 = np.square(v2)
            for i in range(1, len(self.fields)):
                v = self.fields[i].values()
                v1 += np.square(v)
                v2 += v
            return v1 / len(self.fields) - np.square(v2 / len(self.fields))

    def latitudes(self):
        v = [x.latitudes() for x in self.fields]
        return Fieldset._make_2d_array(v)

    def longitudes(self):
        v = [x.longitudes() for x in self.fields]
        return Fieldset._make_2d_array(v)

    def _lat_func(self, func, bitmap_poles=False):
        result = Fieldset(temporary=True)
        pole_limit = 90.0 - 1e-06
        with open(result.temporary.path, "wb") as fout:
            for f in self.fields:
                lat = f.latitudes()
                if bitmap_poles:
                    lat[np.fabs(lat) > pole_limit] = np.nan
                v = func(np.deg2rad(lat))
                c = f.clone()
                c.encode_values(v)
                result._append_field(c)
                result.fields[-1].write(
                    fout, result.temporary.path, temp=result.temporary
                )
        return result

    def coslat(self):
        return self._lat_func(np.cos)

    def sinlat(self):
        return self._lat_func(np.sin)

    def tanlat(self):
        return self._lat_func(np.tan, bitmap_poles=True)

    @staticmethod
    def _list_or_single(lst):
        return lst if len(lst) != 1 else lst[0]

    @staticmethod
    def _always_list(items):
        return items if isinstance(items, list) else [items]

    @staticmethod
    def _make_2d_array(v):
        """Forms a 2D ndarray from a list of 1D ndarrays"""
        v = Fieldset._list_or_single(v)
        return np.stack(v, axis=0) if isinstance(v, list) else v

    # TODO: add all the field_func functions

    # TODO: add all field_field_func functions

    # TODO: add all field_scalar_func functions

    # TODO: add tests for all fieldset-fieldset methods:
    # **, ==, !=, &, |

    # TODO: allow these methods to be called as functions (?)

    # TODO: function to write to single file if fields from different files

    # TODO: optimise write() to copy file if exists

    # TODO: to_dataset()

    # TODO: pickling

    # TODO: gribsetbits, default=24

    def _get_db(self):
        if self._db is None:
            self._db = indexdb.FieldsetDb(fs=self)
        assert self._db is not None
        return self._db

    def _unique_metadata(self, key):
        return self._get_db().unique(key)

    def select(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            return self._get_db().select(**args[0])
        else:
            return self._get_db().select(**kwargs)

    def describe(self, *args, **kwargs):
        return self._get_db().describe(*args, **kwargs)

    def ls(self, **kwargs):
        return self._get_db().ls(**kwargs)

    def sort(self, *args, **kwargs):
        return self._get_db().sort(*args, **kwargs)

    def deacc(self, **kwargs):
        return utils.deacc(self, **kwargs)

    def speed(self, *args):
        if len(args) == 0:
            u = self[0::2]
            v = self[1::2]
            if len(u) != len(v):
                raise Exception(
                    f"Fieldsets must contain an even number of fields for this operation! len={len(self.fields)} is not even!"
                )
            return u.speed(v)
        elif len(args) == 1:
            other = args[0]
            result = Fieldset(temporary=True)
            param_ids = {
                131: 10,  # atmospheric wind
                165: 207,  # 10m wind
                228246: 228249,  # 100m wind
                228239: 228241,  # 200m wind
            }
            if len(self.fields) != len(other):
                raise Exception(
                    f"Fieldsets must have the same number of fields for this operation! {len(self.fields)} != {len(other)}"
                )
            with open(result.temporary.path, "wb") as fout:
                for f, g in zip(self.fields, other.fields):
                    sp = np.sqrt(np.square(f.values()) + np.square(g.values()))
                    c = f.clone()
                    c.encode_values(sp)
                    param_id_u = f.grib_get("paramId", CodesHandle.LONG)
                    param_id_sp = param_ids.get(param_id_u, None)
                    if param_id_sp is not None:
                        c = c.grib_set(["paramId", param_id_sp], CodesHandle.LONG)
                    result._append_field(c)
                    result.fields[-1].write(
                        fout, result.temporary.path, temp=result.temporary
                    )
            return result


class FieldsetCF:
    def __init__(self, fs):
        self.fs = fs

    def items(self):
        its = []
        for i in range(len(self.fs)):
            its.append((i, FieldCF(self.fs[i])))
        # print("FieldsetCF items: ", d)
        return its

    def __getitem__(self, index):
        return FieldCF(self.fs[index])


class FieldCF:

    QUALIFIER_MAP = {"float": "d"}

    INT_KEYS = ["Nx", "Ny", "number"]

    def __init__(self, fs):
        self.fs = fs

    def __getitem__(self, key):
        keyname = key
        qualifier = "n"
        parts = key.split(":")
        if len(parts) > 1:
            keyname = parts[0]
            qualifier = FieldCF.QUALIFIER_MAP[parts[1]]

        native_key = keyname + ":" + qualifier
        # print("native_key", native_key)
        value = self.fs.grib_get([native_key])[0][0]
        # print(value)
        if value is None:
            raise IndexError
        if key in FieldCF.INT_KEYS:
            # print("int value:", int(value))
            return int(value)
        # if isinstance(value, float):
        #    return int(value)
        return value


def read(p):
    return Fieldset(path=p)


# expose all Fieldset functions as a module level function
def _make_module_func(name):
    def wrapped(fs, *args):
        return getattr(fs, name)(*args)

    return wrapped


module_obj = sys.modules[__name__]
for fn in dir(Fieldset):
    if callable(getattr(Fieldset, fn)) and not fn.startswith("_"):
        setattr(module_obj, fn, _make_module_func(fn))


def bind_functions(namespace, module_name=None):
    """Add to the module globals all metview functions except operators like: +, &, etc."""
    namespace["read"] = read
    for fn in dir(Fieldset):
        if callable(getattr(Fieldset, fn)) and not fn.startswith("_"):
            namespace[fn] = _make_module_func(fn)
        namespace["Fieldset"] = Fieldset



# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import metview as mv

# this module is meant to expose some ui specific functions


def dialog(*args):
    res = mv._dialog(*args)
    return {k: v for k, v in res.items() if not k.startswith("_")}


def any(**kwargs):
    return mv._any(**kwargs)


def colour(**kwargs):
    return mv._colour(**kwargs)


def icon(**kwargs):
    return mv._icon(**kwargs)


def option_menu(**kwargs):
    return mv._option_menu(**kwargs)


def slider(**kwargs):
    return mv._slider(**kwargs)


def toggle(**kwargs):
    return mv._toggle(**kwargs)
